import functools
import os

import pytest

from granola import Cereal, SerialSniffer
from granola.utils import IS_PYTHON3

if IS_PYTHON3:
    from unittest.mock import patch
else:
    from mock import patch

current_dir = os.path.dirname(os.path.realpath(__file__))
CONFIG_PATH = os.path.join(current_dir, "config.json")
CONFIG_PATH_DEPRECATIONS = os.path.join(current_dir, "config_deprecations.json")


def assert_filled_all(mask):
    assert not mask.empty
    assert mask.all()


def assert_filled_any(mask):
    assert not mask.empty
    assert mask.any()


def all_equal(iterator):
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == x for x in iterator)


def query_device(bk_cereal, cmd, write_terminator="\r"):
    bk_cereal.write(
        ("{cmd}{write_terminator}".format(cmd=cmd, write_terminator=write_terminator)).encode(bk_cereal._encoding)
    )
    return bk_cereal.read(1000)


def decode_response(response, bk_cereal):
    return response.decode(bk_cereal._encoding)


def check_deprecation(*msgs):
    if callable(msgs[0]):
        func = msgs[0]
        msg = "deprecated"

        @functools.wraps(func)
        def _inner(*args, **kwargs):
            with pytest.deprecated_call() as dep:
                r = func(*args, **kwargs)
                assert issubclass(dep.list[-1].category, Warning)
                assert msg in str(dep.list[-1].message).lower()
            return r

        return _inner
    else:

        def _check_deprecation(func):
            @functools.wraps(func)
            def _inner(*args, **kwargs):
                with pytest.deprecated_call() as dep:
                    r = func(*args, **kwargs)
                    found_warning = False
                    for msg in msgs:
                        for warning in dep.list:
                            if issubclass(warning.category, Warning) and msg.lower() in str(warning.message).lower():
                                found_warning = True
                                break
                        assert found_warning
                return r

            return _inner

        return _check_deprecation


class SerialSnifferTester(SerialSniffer):
    outfile = "test_output.csv"


@pytest.fixture
def mock_cereal():
    return Cereal.mock_from_json("cereal", config_path=CONFIG_PATH)()


@pytest.fixture
def bk_cereal_only_getters_and_setters():
    return Cereal.mock_from_json("just_getters_and_setters", config_path=CONFIG_PATH)()


@pytest.fixture
def sniff_sniff():
    sniff = SerialSnifferTester()
    yield sniff
    os.remove(sniff.outpath)


@pytest.fixture
def mock_write():
    with patch("serial.Serial.write") as write:
        yield write


@pytest.fixture
def mock_read():
    with patch("serial.Serial.read") as read:
        yield read


@pytest.fixture
def canned_queries_command_readers():
    command_readers = {
        "CannedQueries": {
            "data": [
                {
                    "1\r": "1",
                    "2\r": {"response": "2"},
                    "3\r": {"response": "3", "delay": 3},
                    "4\r": {"response": ["4a", "4b"]},
                    "5\r": {"response": ["5a", "5b"], "delay": 5},
                    "6\r": {"response": ["6a", "6b"], "delay": [6.1, 6.2]},
                    "7\r": {"response": [["7a", {"delay": 7}], "7b"]},
                    "8\r": {"response": [["8a", {"delay": 8.1}], "8b"], "delay": 8},
                    "9\r": [["9a", {"delay": 9}], "9b"],
                }
            ],
            "delay": 0,
        }
    }
    return command_readers



import csv
import logging
from builtins import bytes
from builtins import (
    str as unicode,  # python 3 doesn't have a unicode type, and we don't want to override python2 str type
)
from contextlib import contextmanager
from datetime import datetime

from serial import Serial

from granola.utils import (
    IS_PYTHON3,
    add_created_at,
    check_min_package_version,
    decode_bytes,
    encode_escape_char,
    get_path,
    is_terminated_with,
    make_path,
)

logger = logging.getLogger(__name__)


class SerialSniffer(Serial):
    """
    This class is meant to be injected in place of Serial in some code that communicates with a device.
    It will then write out any serial commands sent, and their responses to a CSV file, to be used by GRANOLA.
    This code assumes that the next write after any read will be the "response"
    unpaired reads and writes will be ignored.

    Args: (passed by subclassing SerialSniffer with the appropriate class variable overwritten)
        outfile (str): Path to the file you want to write to
        delimiter : passed to csv.writer()
        quotechar : passed to csv.writer()
        write_terminator (str) : The character sequence used to indicate that a serial command is complete
        read_terminator (str) : The character sequence used to indicate that a serial response is complete

    .. todo::

        Change how this is done to the same way the Breakfast Cereal is done (init then __call__)
    """

    outfile = ""
    delimiter = ","
    quotechar = '"'
    write_terminator = b"\r"
    read_terminator = b"\r>"

    @add_created_at
    def __init__(self, *args, **kwargs):
        logger.debug("%s initializing Sniffer with Serial args %s and kwargs %s", self.__class__.__name__, args, kwargs)
        super(SerialSniffer, self).__init__(*args, **kwargs)
        if self.outfile == "":
            if self.port is not None:
                self.outfile = datetime.now().strftime("%Y-%m-%dT%H-%M-%S") + "_" + self.port + ".csv"
            else:
                self.outfile = datetime.now().strftime("%Y-%m-%dT%H-%M-%S") + "_serial_commands.csv"

        self.outpath = get_path(self.outfile)
        make_path(self.outpath)

        logger.debug("%s outpath: %s", self, self.outpath)

        self.current_write = b""
        self.current_read = b""
        self.last_write_time = datetime.now()
        self.last_read_time = datetime.now()

        try:
            with _open_csv_writer(
                self.outpath, "w", delimiter=self.delimiter, quotechar=self.quotechar, quoting=csv.QUOTE_MINIMAL
            ) as csvwriter:
                csvwriter.writerow(["cmd", "response", "delay(ms)"])
        except IOError as err:
            logger.exception("%s %r", self, err)

    def __str__(self):
        port = getattr(self, "port", "")
        port_str = " on %s" % port if port else ""
        return "{device}{port}".format(device=self.__class__.__name__, port=port_str)

    def write(self, data, *args, **kwargs):
        """
        A wrapper for Serial.write, that also stores written content, and when a terminator is reached,
        records that read to the given csv inputs and outputs are the same as Serial.read
        Note: to mock the behavior of the serial instruments I have access to, any charaters after the initial
        terminator are ignored for our purposes. They are still written to the serial port, of course."""

        logger.info("%s write: %r", self, data)

        if not self._is_write_terminated(self.current_write):
            # this write is not yet terminated
            for int_repr in bytes(data):  # convert byte data into ascii int representation
                self.current_write += bytes([int_repr])  # convert back to bytes
                if self._is_write_terminated(self.current_write):
                    self.last_write_time = datetime.now()
                    self.current_read = b""
                    break

        return super(SerialSniffer, self).write(data, *args, **kwargs)

    def read(self, size=1, *args, **kwargs):
        """
        A wrapper for Serial.read, that also stores read content, and when a terminator is reached,
        records that read to the given csv inputs and outputs are the same as Serial.read"""
        read = super(SerialSniffer, self).read(size=size, *args, **kwargs)

        logger.info("%s read: %r", self, read)

        self.current_read += read

        if self._is_read_terminated(self.current_read):
            self.last_read_time = datetime.now()
            delay = (self.last_read_time - self.last_write_time).total_seconds() * 1000
            try:
                if self._is_write_terminated(self.current_write):
                    with _open_csv_writer(
                        self.outpath, "a", delimiter=self.delimiter, quotechar=self.quotechar, quoting=csv.QUOTE_MINIMAL
                    ) as csvwriter:
                        csvwriter.writerow(
                            [
                                encode_escape_char(decode_bytes(self.current_write)),
                                encode_escape_char(decode_bytes(self.current_read)),
                                delay,
                            ]
                        )
            except IOError as err:
                logger.exception("%s %r", self, err)
            self.current_write = b""
            self.current_read = b""

        return read

    if check_min_package_version("pyserial", "3.0"):

        def reset_input_buffer(self):
            """
            A wrapper for serial.reset_input_buffer that also clears the current read buffer.
            Should only be used with pyserial versions >= 3.0"""
            self.current_read = b""
            super(SerialSniffer, self).reset_input_buffer()

        def reset_output_buffer(self):
            """
            A wrapper for serial.reset_output_buffer that also clears the current write buffer.
            Should only be used with pyserial versions >= 3.0"""
            self.current_write = b""
            super(SerialSniffer, self).reset_output_buffer()

    else:

        def flushInput(self):
            """
            A wrapper for serial.FlushInput that also clears the current read buffer.
            Should only be used with pyserial versions <= 3.0"""
            self.current_read = b""
            super(SerialSniffer, self).flushInput()

        def flushOutput(self):
            """
            A wrapper for serial.FlushOutput that also clears the current write buffer.
            Should only be used with pyserial versions <= 3.0"""
            self.current_write = b""
            super(SerialSniffer, self).flushOutput()

    def _is_write_terminated(self, input):
        return is_terminated_with(input, self.write_terminator)

    def _is_read_terminated(self, input):
        return is_terminated_with(input, self.read_terminator)


@contextmanager
def _open_csv_writer(path, mode, **kwargs):
    """
    A contextmanager to open the csv file at a given path (in either python 2 or 3) and return a writer object
    for that file

    Args:
        path (str): the path to your csv file
        mode (str): the access mode passed on to open
        additional **kwargs are passed to csv.writer()

    Returns:
        csv.writer: a handle for the csv writer
    """

    if IS_PYTHON3:
        csvfile = open(path, mode, newline="")
    else:
        csvfile = open(path, mode + "b")
        kwargs = {
            k: str(v) if isinstance(v, unicode) else v  # python 2 csv writer needs byte strings
            for k, v in kwargs.items()
        }

    csvwriter = csv.writer(csvfile, **kwargs)
    try:
        yield csvwriter
    finally:
        csvfile.close()


__doc__ = """
``SerialSniffer`` is a class that can be used in place of pyserial Serial to track every write and read for pyserial.
All it does is capture every incoming write and read, writes it out to a csv, and then passes it along to
pyserial.

You can easily run ``SerialSniffer`` in your code for a one off to just capture some commands for mocking later
by replacing your pyserial Serial import with::

    from granola import SerialSniffer as Serial
"""



import pandas as pd

from granola import CannedQueries, Cereal, RandomizeResponse
from granola.tests.conftest import (
    CONFIG_PATH,
    all_equal,
    assert_filled_all,
    query_device,
)


def test_that_you_can_pass_canned_queries_directly_instead_of_as_file_paths(canned_queries_command_readers):
    # Given a mock pyserial class defined by a python dictionary configuration with canned queries directly

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    # and query it
    one = query_device(mock, "1")

    # Then we get the expected result
    assert one == b"1"


def test_that_you_can_pass_canned_queries_directly_with_multiple_response(canned_queries_command_readers):
    # Given a mock pyserial class defined by a python dictionary configuration with canned queries directly

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    # and issue get temp twice
    a4 = query_device(mock, "4")
    b4 = query_device(mock, "4")

    # Then we should get different responses
    assert a4 == b"4a"
    assert b4 == b"4b"


def test_that_you_can_pass_canned_queries_directly_with_multiple_commands(canned_queries_command_readers):
    # Given a mock pyserial class defined by a python dictionary configuration with canned queries directly

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    # and issue all of our serial commands
    a4 = query_device(mock, "4")
    b4 = query_device(mock, "4")
    one = query_device(mock, "1")
    a6 = query_device(mock, "6")

    # Then we should the specified responses
    assert a4 == b"4a"
    assert b4 == b"4b"
    assert one == b"1"
    assert a6 == b"6a"


def test_that_you_can_pass_canned_queries_directly_with_multiple_commands_from_json():
    # Given a mock pyserial class defined by a json dictionary configuration with canned queries directly

    # When we initialize it
    mock = Cereal.mock_from_json("dictionary", config_path=CONFIG_PATH)()
    # and issue all of our serial commands
    twenty = query_device(mock, "get -temp")
    twenty_two = query_device(mock, "get -temp")
    sn1234 = query_device(mock, "get -sn")
    ok = query_device(mock, "test -off")

    # Then we should the specified responses
    assert twenty == b"20\r>"
    assert twenty_two == b"22\r>"
    assert sn1234 == b"sn1234\r>"
    assert ok == b"OK\r>"


def test_that_you_can_specify_a_delay_on_one_command():
    # Given a dictionary canned queries with a delay on only one command
    command_readers = {
        "CannedQueries": {
            "data": [{"1\r": "1", "2\r": {"response": "3", "delay": 3}}],
        }
    }

    # When we initialize it
    mock = Cereal(command_readers=command_readers)()
    df = mock._readers_["CannedQueries"].serial_df

    # Then 2 has a delay of 3 but get -sn does not have any delay (nan)
    assert_filled_all(df.loc[(df.cmd == "2\r")]["delay"] == 3)
    assert_filled_all(df.loc[(df.cmd == "1\r")]["delay"].isna())


def test_that_you_can_specify_a_delay_on_one_command_and_a_broadcasting_delay_for_the_rest(
    canned_queries_command_readers,
):
    # Given a dictionary canned queries with a default delay and a specific delay

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    df = mock._readers_["CannedQueries"].serial_df

    # Then 3 gets the default delay of 3, but 2 has a delay of 0 since we that is the default
    assert_filled_all(df.loc[(df.cmd == "3\r")]["delay"] == 3)
    assert_filled_all(df.loc[(df.cmd == "2\r")]["delay"] == 0)


def test_that_you_can_specify_a_inside_a_response_list(canned_queries_command_readers):
    # Given a dictionary canned queries with a default delay and a specific delay

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    df = mock._readers_["CannedQueries"].serial_df

    # THen the delays inside lists should specify columns
    assert_filled_all(df.loc[(df.cmd == "1\r")]["delay"] == 0)
    assert_filled_all(df.loc[(df.cmd == "3\r")]["delay"] == 3)
    assert_filled_all(df.loc[(df.cmd == "7\r") & (df.response == "7a")]["delay"] == 7)
    assert_filled_all(df.loc[(df.cmd == "7\r") & (df.response == "7b")]["delay"] == 0)


def test_that_that_all_off_the_ways_to_specify_canned_queries_inside_dicts_can_go_together(
    canned_queries_command_readers,
):
    # Given a dictionary canned queries with a default delay and a specific delay

    # When we initialize it
    mock = Cereal(command_readers=canned_queries_command_readers)()
    df = mock._readers_["CannedQueries"].serial_df

    # THen we should get get the delays inside lists should specify columns
    assert_filled_all(df.loc[(df.cmd == "1\r")]["delay"] == 0)
    assert_filled_all(df.loc[(df.cmd == "2\r")]["delay"] == 0)
    assert_filled_all(df.loc[(df.cmd == "3\r")]["delay"] == 3)
    assert_filled_all(df.loc[(df.cmd == "4\r")]["delay"] == 0)
    assert_filled_all(df.loc[(df.cmd == "5\r")]["delay"] == 5)
    assert_filled_all(df.loc[(df.cmd == "6\r") & (df.response == "6a")]["delay"] == 6.1)
    assert_filled_all(df.loc[(df.cmd == "6\r") & (df.response == "6b")]["delay"] == 6.2)
    assert_filled_all(df.loc[(df.cmd == "7\r") & (df.response == "7a")]["delay"] == 7)
    assert_filled_all(df.loc[(df.cmd == "7\r") & (df.response == "7b")]["delay"] == 0)
    assert_filled_all(df.loc[(df.cmd == "8\r") & (df.response == "8a")]["delay"] == 8.1)
    assert_filled_all(df.loc[(df.cmd == "8\r") & (df.response == "8b")]["delay"] == 8)
    assert_filled_all(df.loc[(df.cmd == "9\r") & (df.response == "9a")]["delay"] == 9)
    assert_filled_all(df.loc[(df.cmd == "9\r") & (df.response == "9b")]["delay"] == 0)


def test_random_responses():
    # When we have a dataframe and and randomized response enum
    df = pd.DataFrame(data=dict(cmd=[1, 2, 3], response=[1, 2, 3]))
    will_randomize_responses = RandomizeResponse.randomized_w_replacement.name

    # When we randomize our response 100 times
    # (we choose 100, just to be pretty sure that it will give us difference respones, even if luck isn't on our side)
    randomized_responses = []
    for _ in range(100):
        randomized_responses.append(next(CannedQueries._get_generator_from_df(df.copy(), will_randomize_responses)))

    # instead of always getting the 1st response, we should get others as well
    assert len(set(randomized_responses)) != 1


def test_random_responses_with_removal():

    # When we have a dataframe and and randomized response enum
    df = pd.DataFrame(data=dict(cmd=[1, 2, 3], response=[1, 2, 3]))
    will_randomize_responses = RandomizeResponse.randomize_and_remove.name

    # When we randomize our response 100 times
    # (we choose 100, just to be pretty sure that it will give us difference respones, even if luck isn't on our side)
    randomized_responses = []
    for _ in range(100):
        responses = []
        for _ in range(len(df)):
            responses.append(next(CannedQueries._get_generator_from_df(df.copy(), will_randomize_responses)))
        randomized_responses.append(responses)

    # instead of always getting the 1st response, we should get others as well
    for responses in randomized_responses:
        assert len(responses) == 3
    assert not all_equal(randomized_responses)


def test_pass_in_randomize_response_in_canned_queries():
    # Given a CannedQueries config with a randomized responses
    command_readers = {
        "CannedQueries": {"data": [{"1\r": ["1a", "1b", "1c"]}], "will_randomize_responses": "randomized_w_replacement"}
    }

    # When we query the entire list of our 1\r command 100 times
    # (we choose 100, just to be pretty sure that it will give us difference respones, even if luck isn't on our side)
    randomized_responses = []
    for _ in range(100):
        mock = Cereal(command_readers=command_readers)()
        responses = []
        for _ in range(3):
            responses.append(query_device(mock, "1"))
        randomized_responses.append(responses)

    # Then the order the 3 respones is not the same accross all 100 tries
    assert not all_equal(randomized_responses)



from granola.utils import IS_PYTHON3

if IS_PYTHON3:
    from enum import Enum
else:  # pragma: no cover
    from aenum import Enum


def get_attribute_from_enum(maybe_enum, enum_attribute):
    """If the passed in maybe_enum is an enum, return the attribute, else return the maybe_enum

    Args:
        maybe_enum (Enum or Enum key)
        enum_attribute (str): The name of the enum attribute to retrieve
    """
    if isinstance(maybe_enum, Enum):
        return getattr(maybe_enum, enum_attribute)
    return maybe_enum


def validate_enum(value, enum):
    """Validate the given value is in the given enum

    Args:
        value (Enum or enum key)
        enum: the enum to validate against
    """
    error = ValueError("Invalid Enum Option\nEnum -- %s\nOption supplied -- %s" % (enum.str_enum_options(), value))
    if isinstance(value, Enum):
        if value not in enum:
            raise error
    elif value not in enum.__members__:
        raise error


class DocumentedEnum(Enum):
    """
    Enum Base class to inherit from that replaces the value option with a description option.

    Examples:
        >>> class HookTypes(DocumentedEnum):
        ...     pre_reading = "Hook to run before `get_reading` methods"
        ...     post_reading = "Hook to run after `get_reading` methods"

        >>> HookTypes.pre_reading.name
        'pre_reading'

        >>> HookTypes.post_reading.description
        'Hook to run after `get_reading` methods'
    """

    def __init__(self, description):
        self.description = description

    @classmethod
    def str_enum_options(cls):
        string = "{enum}\n".format(enum=cls.__name__)
        nl = "\n"
        string += nl.join(
            "Option: {name} -- Description: {doc}".format(name=enum.name, doc=enum.description) for enum in cls
        )
        return string


class RandomizeResponse(DocumentedEnum):
    """
    The different ways to specify to randomize or not randomize your
    :class:`~granola.command_readers.CannedQueries` response
    """

    not_randomized = "Don't randomize the response"
    randomized_w_replacement = "Randomize with replacement"
    randomize_and_remove = "Randomize and remove"


class HookTypes(DocumentedEnum):
    """
    Allowed hook types for ``BaseHook`` methods or for a ``register_hook``
    Current allowed hook types are ``pre_reading`` and ``post_reading``.
    """

    pre_reading = "Hook to run before `get_reading` methods"
    post_reading = "Hook to run after `get_reading` methods"


class SetRelationship(DocumentedEnum):
    """
    The different ways to specify what a hook should do with commands or attributes,
    should the hook ``include`` everything listed in the corresponding data container
    and nothing else, or should it exclude everything in the corresponding data container
    and include everything else.
    """

    include = "Include everything listed in the corresponding data container"
    exclude = "Exclude everything listed in the corresponding data container"



import functools
import inspect

from granola.enums import (
    HookTypes,
    SetRelationship,
    get_attribute_from_enum,
    validate_enum,
)
from granola.utils import ABC, IS_PYTHON3


class BaseHook(ABC):
    hooked_classes = []

    def __init__(self, attributes=None, include_or_exclude=SetRelationship.exclude, *args, **kwargs):
        self.attributes = attributes if attributes is not None else {}
        self.include_or_exclude = get_attribute_from_enum(maybe_enum=include_or_exclude, enum_attribute="name")
        validate_enum(self.include_or_exclude, SetRelationship)

    def pre_reading(self, hooked, data, **kwargs):
        """
        Run the pre_reading hook

        Args:
            hooked: instance of hooked class
            data (str): Serial command

        Returns:
            data (str)
        """
        return data

    def post_reading(self, hooked, result, data, **kwargs):
        """
        Run the post_reading hook

        Args:
            hooked: instance of hooked class
            result (str): Returned result from serial command.
            data (str): Serial command

        Returns:
            result (str)
        """
        return result


def register_hook(hook_type_enum, hooked_classes):  # TODO madeline allow passing in attributes and including
    """
    Register a function as a `BaseHook` subclass (this converts a function into a subclass of
    `BaseHook`. `hook_type` is the method(s) to insert your function as in the created subclass.
    (example, if you say `hook_types="pre_reading"`, then the decorated function will become the
    `pre_reading` method in the returned Hook class.

    Options for `hook_type` are any enum in `HookTypes`. `hooked_classes` are the classes that you
    wish to run this hook on.

    Args:
        hook_type (str | HookTypes): str representation of `HookTypes` or `HookTypes` instances themselves
            you wish to run this hook as. Your option(s) will be coverted to methods in the returned class.
        hooked_classes (list[BaseCommandReaders]): list of `BaseCommandReaders` that you wish to
            run this hook on.
    """
    hook_classes = hooked_classes

    def _register_hook(func):
        class RegisteredHook(BaseHook):
            hooked_classes = hook_classes

        hook_type = get_attribute_from_enum(hook_type_enum, "name")
        validate_enum(hook_type, HookTypes)
        if IS_PYTHON3:  # pragma: no cover
            inspection_getargspec = inspect.getfullargspec
        else:
            inspection_getargspec = inspect.getargspec
        func_args = inspection_getargspec(func).args  # grab function signature

        def hook(self, **kwargs):
            in_attribs = kwargs.get("data") in self.attributes
            include_attribs = self.include_or_exclude == SetRelationship.include.name
            exclude_attribs = self.include_or_exclude == SetRelationship.exclude.name
            # Only run if data is to be included or not excluded
            if (in_attribs and include_attribs) or (not in_attribs and exclude_attribs):

                # extract out the params of func from kwargs
                kws = {k: v for k, v in kwargs.items() if k in func_args}
                if "self" in func_args:
                    kws["self"] = self  # only pass in self if self in func signature
                return func(**kws)
            if "result" in kwargs:
                return kwargs["result"]

        # Copy over original function meta data to new class (in desired place, so functools doesn't work exactly)
        hook.__doc__ = func.__doc__
        RegisteredHook.__module__ = func.__module__
        RegisteredHook.__name__ = func.__name__

        setattr(RegisteredHook, hook_type, hook)

        return RegisteredHook

    return _register_hook


def _run_pre_reading_hooks(hooked, data, **kwargs):
    for hook in hooked._hooks_:
        data = hook.pre_reading(hooked=hooked, data=data, **kwargs)
    return data


def _run_post_reading_hooks(hooked, result, data, **kwargs):
    for hook in hooked._hooks_:
        result = hook.post_reading(hooked=hooked, result=result, data=data, **kwargs)
    return result


def wrap_in_hooks(func):
    """
    Decorator that wraps a function in hooks that will run pre and post function.

    wrapper args:
        hooked (CannedQueries): CannedQueries instance to run the hook on.
        result (str): Passed in result to potentially modify.
        data (str): Serial command.
    """

    @functools.wraps(func)
    def wrapper(hooked, data, **kwargs):
        data = _run_pre_reading_hooks(hooked=hooked, data=data, **kwargs)
        result = func(hooked, data, **kwargs)
        result = _run_post_reading_hooks(hooked=hooked, result=result, data=data, **kwargs)
        return result

    return wrapper


__doc__ = r"""
This module provides the abstract base class for Hook objects, ``BaseHook``.

``BaseHook`` outlines the framework to design a hook class.

* You inherit from ``BaseHook``
* specify your ``hooked_classes``, which is a list of all the classes that your hook will run on.
* If you need additional initialization arguments, you can override the default init, passing
  ``attributes`` and ``include_or_exclude`` to ``BaseHook`` init.
* Finally, define what hook methods you need. That might be running hooks ``pre_reading`` or
  ``post_reading`` or both

* You can also define a hook as a function instead of a class and then "register" it as a hook
  with the ``register_hook`` decorator below. This can be very efficient for simple hooks,
  but it does lose you the ability to run both a ``pre_reading`` and ``post_reading`` method
  on the same hook, which more complicated hooks maybe need.

****************
Hooks as Classes
****************

Here is an example of a hook class being created

>>> from granola import GettersAndSetters, CannedQueries, BaseHook

>>> class MyHook(BaseHook):
...     hooked_classes=[GettersAndSetters, CannedQueries]
...     '''This is the list of classes that the hook will be run on.'''
...
...     def __init__(self, attributes=["get -sn\r"], include_or_exclude="exclude", some_parameter="default_val"):
...         '''attributes and include_or_exclude are parameters (with defaults) that come from
...         BaseHook. This specifies which serial commands or other type of attribute will be
...         include or exclude on your hook (which will be ran). By saying attributes=["get -sn\r"]
...         and include_or_exclude="include" we are saying to exclude "get -sn\r" and
...         include all other commands.
...         '''
...
...         super(MyHook, self).__init__(attributes=attributes, include_or_exclude=include_or_exclude)
...         self.some_parameter = some_parameter
...         self.calculated_value = None
...
...     def pre_reading(self, hooked, data, **kwargs):
...         '''store the length of the the incoming command (data) * ``self.some_parameter``'''
...         self.calculated_value = self.some_parameter * len(data)
...
...     def post_reading(self, hooked, result, data, **kwargs):
...         '''Change the serial result to the previous ``self.calculated_value`` * ``len(result)``'''
...         result = self.calculated_value * len(result)
...         return result

Doing a hook this way, as a class definition allows you to specify a ``pre_reading`` and ``post_reading``
method on the same hook. Some hooks (such as the approach hook) rely on it being ran before and after the reading
to know the state before the actual reading, and to change the result for some commands.

******************
Hooks as Functions
******************

If your hook doesn't require both a ``pre_reading`` and ``post_reading``, you can instead just write
a function and decorate it with the :func:`~granola.hooks.base_hook.register_hook` decorator.

Here is an example of a hook being created

>>> from granola import register_hook, CannedQueries

>>> @register_hook(hook_type_enum="post_reading", hooked_classes=[CannedQueries])
... def LoopCannedQueries(hooked, result, data, **kwargs):
...     if result is SENTINEL:
...         hooked._start_serial_generator(data)
...         result = next(hooked.serial_generator[data])
...         return result
...     return result

This is the actual implementation of :class:`~granola.hooks.hooks.LoopCannedQueries` the looping hook that by
default loops all canned queries (minus the docstring)

:func:`~granola.hooks.base_hook.register_hook` takes two arguments, the ``hook_type_enum``
(see :class:`~granola.enums.HookTypes` for options and more details), which tells you if it is a
pre or post_reading hook (can be passed in as a direct enum object or as a string), and the ``hooked_classes``,
which tell you which :mod:`~granola.command_readers` the hook applies to.
After that it is just your function definition.

The signature of your function should match the signature of base method you are using
(``post_reading``, ``post_reading``, etc.) from :class:`~granola.hooks.base_hook.BaseHook`.

After you have "registered" your hook function, it can then be initialized like a hook class.
(i.e. substantiating it before you pass it to ``Cereal``). If you don't instantiate it, then Cereal
will instantiate it with default arguments, this allows you to treat it as a regular function being passed
into ``Cereal``, and not worry about what else happens with the decorator

******************
Hook Examples
******************

>>> from granola import LoopCannedQueries, Cereal
>>> command_readers = {"CannedQueries": {"data": [{"cmd\r": "response"}]}}

>>> # this way works by initializing LoopCannedQueries to non default values
>>> hook = LoopCannedQueries(attributes={"get batt\r"}, include_or_exclude="include")
>>> cereal = Cereal(command_readers=command_readers, hooks=[hook])

>>> # This version we initialize it to its default values
>>> hook = LoopCannedQueries()
>>> ccereal = Cereal(command_readers=command_readers, hooks=[hook])

>>> # This version we don't initialize it at all and let Cereal initialize it to its default values
>>> hook = LoopCannedQueries
>>> cereal = Cereal(command_readers=command_readers, hooks=[hook])

These examples also shows how to pass hooks to ``Cereal``. You pass hooks in an iterable to
``Cereal``, and ``Cereal`` will associate the hook with whatever class it needs to
based on the classes listed in your hook's ``hooked_classes``. In these example, ``LoopCannedQueries`` would
only run on ``CannedQueries``, but ``MyHook`` would run on ``CannedQueries`` and ``GettersAndSetters``.

One final note, SENTINEL is a special book keeping object for special unhandled responses that is different
from a None response.
"""



import pytest

from granola.tests.conftest import query_device


def test_bk_cereal_should_not_have_methods_that_arent_in_pyserial(mock_cereal):
    # Given a mock pyserial class has been initialized

    with pytest.raises(AttributeError):
        # When we access a method implemented in not implemented  in pyserial and not in bk_cereal
        mock_cereal.some_random_method_name_that_doesnt_exist()
        # Then it should raise an error


def test_bk_cereal_should_default_to_configjson_in_same_directory_as_class_definition(mock_cereal):
    # Given a mock pyserial class has been defined in a directory with a config.json and not told about it

    # When we try and access anything from that mock serial
    query_device(mock_cereal, "get name")

    # It should be fine, and nothing should error out


def test_bk_cereal_should_raise_if_writing_with_unicode_strings(mock_cereal):
    # Given a mock pyserial class has been initialized and a unicode input
    input = u"get name\r"

    # When we write to the mock serial with this unicode string
    with pytest.raises(TypeError):
        mock_cereal.write(input)
        # Then it should raise a TypeError



"""
This module is deprecated and will be removed in version 1.0. Please use  :mod:`~granola.breakfast_cereal` instead
"""
from granola.breakfast_cereal import Cereal
from granola.utils import deprecation


class MockSerial(Cereal):
    """
    Deprecated version of :class:`~granola.breakfast_cereal.Cereal`. Will be removed in 1.0

    Switch to using :class:`~granola.breakfast_cereal.Cereal` and the new interface.
    """

    def __init__(self, config_key, config_path="config.json", command_readers=None, hooks=None):
        deprecation("MockSerial is deprecated. Please use granola.breakfast_cereal.Cereal instead.", "1.0")
        command_readers = command_readers or []
        hooks = hooks or []
        config = self._load_json_config(config_key=config_key, config_path=config_path)

        config = self._check_and_normalize_config_deprecation(config)

        config = self._check_and_normalize_command_readers_deprecations(config, command_readers)
        config = self._check_and_normalize_hook_deprecations(config, hooks)

        super(MockSerial, self).__init__(data_path_root=config_path, **config)

    def _check_and_normalize_config_deprecation(self, config):
        command_readers = config.setdefault("command_readers", {})
        if "canned_queries" in config:
            self._check_and_normalize_canned_queries_deprecation(config, command_readers)

        if "getters_and_setters" in config:
            self._check_and_normalize_getters_and_setters_deprecation(config, command_readers)

        return config

    def _check_and_normalize_getters_and_setters_deprecation(self, config, command_readers):
        deprecation(
            "Specifically GettersAnd_setters Command Reader through the outermost config key"
            " 'getters_and_setters is deprecated. Please use the 'command_reader' section instead."
            " See https://granola.readthedocs.io/en/latest/config/config.html for more details."
        )

        # Check for old form of variable substitution pre jinja
        getters_and_setters = config["getters_and_setters"]
        start_not_in = "variable_start_string" not in getters_and_setters
        end_not_in = "variable_end_string" not in getters_and_setters
        if start_not_in and end_not_in:
            deprecation(
                "'GettersAndSetters' variable declaration follows old format"
                "\nSwitch to using ``Cereal``, which defaults to traditional jinja2 formatting ({{ var }}),"
                "\nor specify explicitly your variable_start_string and variable_end_string inside"
                " getters and setters (ex: 'variable_start_string': '`')",
                "1.0",
            )
            # specify getters and setters variable start and end as the old way
            getters_and_setters["variable_start_string"] = "`"
            getters_and_setters["variable_end_string"] = "`"

            # swap canned_queries for "command_readers": {"CannedQueries": ...}
        command_readers["GettersAndSetters"] = config.pop("getters_and_setters")

        getters = command_readers["GettersAndSetters"].get("getters", [])
        for getter in getters:
            if "getter" in getter:
                deprecation(
                    "Using 'getter' key inside"
                    " config['getters_and_setters']['getters']['getter']"
                    "is deprecated and will be removed in a future release."
                    "\nSwitch to using the key 'cmd' instead.",
                    "1.0",
                )
                # swap getters key for cmd
                getter["cmd"] = getter.pop("getter")
        setters = command_readers["GettersAndSetters"].get("setters", [])
        for setter in setters:
            if "setter" in setter:
                deprecation(
                    "Using 'setter' key inside "
                    " config['getters_and_setters']['setters']['setter']"
                    "is deprecated and will be removed in a future release."
                    "\nSwitch to using the key 'cmd' instead.",
                    "1.0",
                )
                # swap setters key for cmd
                setter["cmd"] = setter.pop("setter")

    def _check_and_normalize_canned_queries_deprecation(self, config, command_readers):
        deprecation(
            "Specifically CannedQueries Command Reader through the outermost config key 'canned_queries"
            " is deprecated. Please use the 'command_reader' section instead."
            " See https://granola.readthedocs.io/en/latest/config/config.html for more details.",
            "1.0",
        )
        # swap canned_queries for "command_readers": {"CannedQueries": ...}
        command_readers["CannedQueries"] = config.pop("canned_queries")

        cr = command_readers["CannedQueries"]

        if "files" in cr:
            deprecation("canned_queries key 'files' has been deprecated. Use the key 'data' instead.", "1.0")
            # swap file key for data
            cr["data"] = cr.pop("files")

        data = cr.get("data", [])
        if isinstance(data, dict):
            deprecation(
                "canned_queries['data'] as a dictionary has been deprecated."
                "\nEither use a list of files or list of dictionaries of cmds and responses instead"
                "See configuration section of documentation.",
                "1.0",
            )
            # Turn dictionary of keys mapped to their values into just a list
            new_data = [value for value in data.values()]
            cr["data"] = new_data

    def _check_and_normalize_command_readers_deprecations(self, config, command_readers):
        config_readers = config["command_readers"]
        for command_reader in command_readers:
            str_not_in = command_reader not in config_readers
            obj_not_in = getattr(command_reader, "__name__", "") not in config_readers
            cls_not_in = command_reader.__class__.__name__ not in config_readers
            if str_not_in and obj_not_in and cls_not_in:
                config["command_readers"][command_reader] = {}
        return config

    def _check_and_normalize_hook_deprecations(self, config, hooks):
        config_hooks = config.setdefault("hooks", {})
        for hook in hooks:
            str_not_in = hook not in config_hooks
            obj_not_in = getattr(hook, "__name__", "") not in config_hooks
            cls_not_in = hook.__class__.__name__ not in config_hooks
            if str_not_in and obj_not_in and cls_not_in:
                config["hooks"][hook] = {}
        return config



import logging
import os
from pathlib import Path

from granola import (
    BaseCommandReaders,
    CannedQueries,
    Cereal,
    GettersAndSetters,
    HookTypes,
    SerialSniffer,
    register_hook,
)
from granola.tests.conftest import CONFIG_PATH, decode_response, query_device
from granola.utils import IS_PYTHON3

if IS_PYTHON3:
    from unittest.mock import patch
else:
    from mock import patch


def test_a_bk_cereal_on_COM1_should_print_that_it_on_COM1():
    # Given a mock pyserial class initialized on COM1 and a sniffer on COM2
    com1 = "COM1"
    com2 = "COM2"

    # When we initialize them and get their __str__s
    bk_cereal = Cereal.mock_from_json("cereal", config_path=CONFIG_PATH)(com1)
    mock_str = str(bk_cereal)
    try:
        with patch("serial.Serial.open"):
            sniffer = SerialSniffer(com2)
        sniffer_str = str(sniffer)
    finally:
        # and we cleanup the created sniffer file
        os.remove(sniffer.outpath)

    # Then it should tell us its class name on its com port
    true_mock_string = "Cereal on COM1"
    true_sniffer_string = "SerialSniffer on COM2"
    assert true_mock_string == mock_str
    assert true_sniffer_string == sniffer_str


def test_a_bk_cereal_with_no_canned_queries_should_be_fine():
    # Given a mock pyserial class defined with no canned queries in its keys

    # When we initialize it
    Cereal.mock_from_json("just_getters_and_setters", config_path=CONFIG_PATH)

    # Then it shouldn't throw any errors


def test_that_a_config_paths_work_when_cwd_is_changed():
    # Given a path to a config that is ina totally different working tree
    cwd = os.getcwd()
    try:
        os.chdir("docs")
        config_path = "../config.json"

        # When we initialize a device with canned queries
        fake_device = Cereal.mock_from_json(config_key="fake device", config_path=config_path)

        # Then The canned queries are loaded from the path relative to the config, not CWD
        assert query_device(fake_device, "reset") == b"OK\r>"

    finally:
        # and we move back to the root directory (so if this test fails we don't break a million other tests)
        os.chdir(cwd)


def test_a_bk_cereal_with_on_GettersAndSetters_command_readers_should_still_work_with_those_commands():
    # Given a mock pyserial class defined with just a GettersAndSetters command_readers
    command_readers = [GettersAndSetters]
    new_sn = "2.718"

    # When we initialize it
    mock = Cereal.mock_from_json("cereal", config_path=CONFIG_PATH, command_readers=command_readers)()
    # and issue get and set sn commands
    query_device(mock, "set -sn %s" % new_sn)
    sn = query_device(mock, "get -sn")

    # Then the serial number should be what we set it to
    true_response = new_sn + "\r>"
    decoded_sn = decode_response(sn, mock)
    assert true_response == decoded_sn


def test_a_bk_cereal_with_on_GettersAndSetters_command_readers_should_throw_error_on_bad_query():
    # Given a mock pyserial class defined with just a GettersAndSetters command_readers
    command_readers = [GettersAndSetters]

    # When we initialize it
    mock = Cereal.mock_from_json("cereal", config_path=CONFIG_PATH, command_readers=command_readers)()
    # and it is passed a query not defined
    response = query_device(mock, "dummy command")

    # Then it should return with "ERROR\r"
    decoded_response = decode_response(response, mock)
    assert "ERROR\r>" == decoded_response


def test_a_bk_cereal_with_no_command_readers_should_still_be_able_to_throw_an_unsupported_response():
    # Given a mock pyserial class defined with just a GettersAndSetters command_readers
    command_readers = [GettersAndSetters]

    # When we initialize it
    mock = Cereal.mock_from_json("cereal", config_path=CONFIG_PATH, command_readers=command_readers)()
    # and it is passed a query not defined
    response = query_device(mock, "dummy command")

    # Then it should return with "ERROR\r"
    decoded_response = decode_response(response, mock)
    assert "ERROR\r>" == decoded_response


def test_that_a_python_dictionary_config_is_just_as_good_as_json():
    # Given a mock pyserial class defined by a python dictionary configuration
    command_readers = [CannedQueries]
    command_readers = {
        CannedQueries: {
            "data": [
                Path(__file__).resolve().parent / "data/cereal_cmds.csv",
            ]
        }
    }

    # When we initialize it
    mock = Cereal(command_readers=command_readers)()
    # and issue test off
    ok = query_device(mock, "reset")

    # Then we should get back OK
    true_response = "OK\r>"
    decoded_sn = decode_response(ok, mock)
    assert true_response == decoded_sn


def test_that_a_python_dictionary_config_lets_you_use_getters_and_setters():
    # Given a mock pyserial class defined by a python dictionary configuration of getters and setters
    command_readers = {
        "GettersAndSetters": {
            "default_values": {"sn": "42"},
            "getters": [{"cmd": "get -sn\n", "response": "{{ sn }}\n>"}],
            "setters": [{"cmd": "set -sn {{ sn }}\n", "response": "OK\n>"}],
        }
    }
    new_sn = "2.718"

    # When we initialize it
    mock = Cereal(command_readers=command_readers, write_terminator="\n")()
    # and issue get and set sn commands
    query_device(mock, "set -sn %s" % new_sn, write_terminator="\n")
    sn = query_device(mock, "get -sn", write_terminator="\n")

    # Then the serial number should be what we set it to
    true_response = new_sn + "\n>"
    decoded_sn = decode_response(sn, mock)
    assert true_response == decoded_sn


def test_a_json_config_can_specify_a_hook___stick_hook_specifically():
    # Given a mock serial with a json config that specifies a stick hook
    # defined to stick on the cmd 1\r
    bk_cereal = Cereal.mock_from_json("stick_hook", config_path=CONFIG_PATH)

    # When we query it enough to ensure it has exhausted the generator
    for _ in range(50):
        bk_cereal.write(b"1\n")
        bk_cereal.read(100)

    responses = set()
    # Then all future queries should be the same
    for _ in range(10):
        bk_cereal.write(b"1\n")
        responses.add(bk_cereal.read(100))

        assert len(responses) == 1
        # and we get back the expected response
        assert b"1" in responses


def test_a_json_config_can_specify_a_hook_arguments_to_exclude_a_query___stick_hook_specifically(caplog):
    # Given a mock serial with a json config that specifies a stick hook, but to exclude 2\r
    bk_cereal = Cereal.mock_from_json("stick_hook", config_path=CONFIG_PATH)

    # When we query a our excluded response past the generator
    responses = []
    for _ in range(10):
        with caplog.at_level(logging.WARNING):
            bk_cereal.write(b"2\n")
            responses.append(bk_cereal.read(100))

    # Then we capture in our log that there was an unhandled response from the hook
    assert "unhandled response return from hooks" in caplog.text
    # and we get back the responses as an Unsupported\r> response
    assert responses[-1] == b"Unsupported\r>"


def test_a_json_config_can_specify_command_readers_as_well():
    # Given a mock serial with a json config that specifies a CannedQueries command reader
    mock = Cereal.mock_from_json("command_readers", config_path=CONFIG_PATH)

    # When we issue a number of commands
    ok1 = query_device(mock, "reset")
    one_hund = query_device(mock, "get batt")
    ok2 = query_device(mock, "start")

    # Then none of them are unsupported
    assert ok1 == b"OK\r>"
    assert one_hund == b"100\r>"
    assert ok2 == b"OK\r>"


def test_that_you_can_pass_in_a_custom_hook_through_the_config_dict_as_a_str_and_obj():
    # Given custom hooks
    @register_hook(hook_type_enum=HookTypes.post_reading, hooked_classes=[CannedQueries])
    def hook1(self, hooked, result, data, **kwargs):
        return "01"

    @register_hook(hook_type_enum=HookTypes.post_reading, hooked_classes=[CannedQueries])
    def hook2(self, hooked, result, data, **kwargs):
        return "02"

    # and a config dictionary that references hooks by name and by the actual object
    hooks = {
        "hook1": {"attributes": ["3\r"], "include_or_exclude": "include"},
        hook2: {"attributes": ["4\r"], "include_or_exclude": "include"},
    }

    # When we initialize it and issue our serial command
    mock = Cereal(hooks=hooks)()
    zero1 = query_device(mock, "3")
    zero2 = query_device(mock, "4")

    # Then the hooks returns 01 and 02
    assert zero1 == b"01"
    assert zero2 == b"02"


def test_that_you_can_pass_in_a_custom_command_reader_through_the_config_dict_as_a_str_and_obj():
    # Given custom command readers
    class CommandReader1(BaseCommandReaders):
        def get_reading(self, data):
            if "1" in data:
                return "Command Reader1"

    class CommandReader2(BaseCommandReaders):
        def get_reading(self, data):
            if "2" in data:
                return "Command Reader2"

    # and a config dictionary that references the command readers by name and by the actual object
    command_readers = ["CommandReader1", CommandReader2]

    # When we initialize it and issue our serial command
    mock = Cereal(command_readers=command_readers)()
    cr1 = query_device(mock, "1")
    cr2 = query_device(mock, "2")

    # Then the hooks returns b"Command Reader1" and "Command Reader2"
    assert cr1 == b"Command Reader1"
    assert cr2 == b"Command Reader2"



from granola._version import get_versions
from granola.breakfast_cereal import Cereal, PortNotOpenError
from granola.command_readers import (
    BaseCommandReaders,
    CannedQueries,
    GettersAndSetters,
    RandomizeResponse,
    SerialCmds,
)
from granola.enums import HookTypes, SetRelationship
from granola.hooks.base_hook import BaseHook
from granola.hooks.hooks import (
    ApproachHook,
    LoopCannedQueries,
    StickCannedQueries,
    register_hook,
)
from granola.main import MockSerial  # deprecated
from granola.serial_sniffer import SerialSniffer

__version__ = get_versions()["version"]
del get_versions


__all__ = [
    "__version__",
    "RandomizeResponse",
    "BaseCommandReaders",
    "SerialCmds",
    "Cereal",
    "MockSerial",  # deprecated
    "PortNotOpenError",
    "GettersAndSetters",
    "CannedQueries",
    "SerialSniffer",
    "BaseHook",
    "ApproachHook",
    "LoopCannedQueries",
    "StickCannedQueries",
    "SetRelationship",
    "HookTypes",
    "register_hook",
]



import os
import shutil

import pytest

from granola.utils import IS_PYTHON3, fixpath, get_path, make_path

if IS_PYTHON3:
    from unittest.mock import patch
else:
    from mock import patch


PATHS_TO_TEST = [
    "a\\b\\c\\d.json",
    "a/b/c/d.json",
    r"a\b\c\d.json",
    r"a/b\c/d.json",
    "a/b\\c\\\\d.json",
    "a/b\\c\\\\d.json",
]


ROOT_PATHS_TO_TEST = ["x\\y\\z", "x/y/z", r"x\y\z", r"x\y/z", "x//y\\\\z"]


def test_using_get_path_with_just_a_file_should_return_absolute_path_to_root_to_file():
    # Given a file name
    file = "test.csv"

    # When we call `get_path` on it
    path = get_path(file)

    # Then we should get back the file name located at our root directory
    cwd = os.getcwd()
    true_path = os.path.join(cwd, file)
    assert true_path == path


def test_using_make_path_with_directories_that_dont_exist_should_create_those_directories():
    # Given a path to a file name
    path_to_file = os.path.join("Toby", "Tali", "test.csv")

    # When we call `make_path` on it
    make_path(path_to_file)

    try:
        # Then it should have created the directories Toby and Tali
        dir_name = os.path.dirname(path_to_file)
        assert os.path.isdir(dir_name)

    finally:
        # cleanup by removing created directories
        shutil.rmtree("Toby")


@pytest.mark.parametrize("path", PATHS_TO_TEST)
@pytest.mark.parametrize(
    "os_name,pathlib_flavor,expected",
    [("nt", "_WindowsFlavour", r"a\b\c\d.json"), ("posix", "_PosixFlavour", r"a/b/c/d.json")],
)
def test_fixpath_normalizes_paths_on_both_windows_and_posix_systems(path, os_name, pathlib_flavor, expected):
    # Given a series of paths in string formats to normalize

    # When we fix their paths on both windows and linux
    with patch("os.name", os_name), patch("pathlib." + pathlib_flavor + ".is_supported", True):
        assert fixpath(path) == expected
    # Then fixpath normalizes the path to the native path style of that os


@pytest.mark.parametrize("path", PATHS_TO_TEST)
@pytest.mark.parametrize("root", ROOT_PATHS_TO_TEST)
@pytest.mark.parametrize(
    "os_name,pathlib_flavor,expected",
    [("nt", "_WindowsFlavour", r"x\y\z\a\b\c\d.json"), ("posix", "_PosixFlavour", r"x/y/z/a/b/c/d.json")],
)
def test_fixpath_takes_joined_paths_on_both_windows_and_posix_systems_and_normalizes_them(
    path, root, os_name, pathlib_flavor, expected
):
    # Given a series of paths and roots in string formats to join and then normalize

    # When we join their paths and then fix their paths on both windows and linux
    with patch("os.name", os_name), patch("pathlib." + pathlib_flavor + ".is_supported", True):
        joined_path = os.path.join(root, path)
        assert fixpath(joined_path) == expected
    # Then their paths are joined and then normalized to the native path style of that os



import pytest

from granola.enums import SetRelationship, get_attribute_from_enum, validate_enum


def test_get_attribute_from_enum_should_return_the_passed_in_enum_value():
    # Given an enum and an attribue we want
    enum = SetRelationship.include
    attribue = "name"

    # When we get that attribute with get_attribute_from_enum
    name = get_attribute_from_enum(enum, attribue)

    # Then it is the name of that attribute
    true_name = "include"
    assert true_name == name


def test_get_attribute_from_enum_should_a_str_when_passed_a_str():
    # Given an string
    string = "some string"
    attribue = "name"

    # When we call get_attribute_from_enum with the string
    name = get_attribute_from_enum(string, attribue)

    # Then we get back the string we passed in
    assert string == name


def test_validate_enum_should_validate_a_correct_enum_type_value():
    # Given an enum and one of its values as an enum type
    enum = SetRelationship
    value = SetRelationship.include

    # When we validate that the enum is in fact of that enum type
    validate_enum(value, enum)

    # Then no errors should be raised


def test_validate_enum_should_validate_an_correct_str_type_value():
    # Given an enum and one of its values as an string
    enum = SetRelationship
    value = "include"

    # When we validate that the enum is in fact of that enum type
    validate_enum(value, enum)

    # Then no errors should be raised


def test_validate_enum_should_raise_an_ValueError_on_incorrect_value():
    # Given an enum and an incorrect value
    enum = SetRelationship
    value = "Some incorrect value"

    # When we validate that the enum is in fact of that enum type
    with pytest.raises(ValueError):
        validate_enum(value, enum)
        # Then it should raise a ValueError



import time

from granola import (
    ApproachHook,
    BaseCommandReaders,
    CannedQueries,
    MockSerial,
    register_hook,
)
from granola.tests.conftest import (
    CONFIG_PATH_DEPRECATIONS,
    check_deprecation,
    decode_response,
    query_device,
)


@check_deprecation("MockSerial is deprecated. Please use granola.breakfast_cereal.Cereal instead")
def test_mock_serial_deprecation():
    # Given 2 mock serials using deprecated config_key style instantiation
    mock1 = MockSerial("deprecated", config_path=CONFIG_PATH_DEPRECATIONS)
    mock2 = MockSerial(config_key="deprecated", config_path=CONFIG_PATH_DEPRECATIONS)

    # When we set the serial number
    for new_sn, mock in enumerate([mock1, mock2]):
        new_sn = str(new_sn)
        query_device(mock, "set -sn %s" % new_sn)
        sn = query_device(mock, "get -sn")

        # Then it should still work even though it has been deprecated
        decoded_sn = decode_response(sn, mock)
        assert new_sn + "\r>" == decoded_sn


@check_deprecation(
    "canned_queries['data'] as a dictionary has been deprecated."
    "\nEither use a list of files or list of dictionaries of cmds and responses instead"
)
def test_default_df_key_deprecation():
    # Given a mock serial with deprecated "files" and "_default_csv_" keys
    mock = MockSerial("deprecated", CONFIG_PATH_DEPRECATIONS)

    fortytwo = query_device(mock, "get -sn")

    # Then it should still work as expected
    assert fortytwo == b"42\r>"


@check_deprecation("'GettersAndSetters' variable declaration follows old format")
def test_jinja2_variable_start_and_end_deprecation():
    # Given a mock serial that passes config path and key in directly and old pre-jinja variable identifiers

    # When we initialize it, internally the variable identifiers are remapped to jinja ones
    mock = MockSerial(config_key="deprecated", config_path=CONFIG_PATH_DEPRECATIONS)

    sn = query_device(mock, "get -sn")

    # Then it should still work as expected
    assert sn == b"42\r>"


@check_deprecation("Using 'getter' key inside", "Using 'setter' key inside")
def test_getter_and_setter_keys_deprecation():
    # Given a mock serial with "getter" and "setter" keys
    mock = MockSerial(config_key="deprecated", config_path=CONFIG_PATH_DEPRECATIONS)

    sn = query_device(mock, "get -sn")

    # Then it should still work as expected
    assert sn == b"42\r>"


def test_passing_in_command_readers_deprecation():
    # Given a mock serial with 2 command readers, one of them custom
    class CommandReader(BaseCommandReaders):
        def get_reading(self, data):
            if "1" in data:
                return "Command Reader"

    mock = MockSerial(
        config_key="deprecated", config_path=CONFIG_PATH_DEPRECATIONS, command_readers=[CannedQueries, CommandReader]
    )

    sn = query_device(mock, "get -sn")
    one = query_device(mock, "1")

    # Then it should trigger our 2 different command readers
    assert sn == b"42\r>"
    assert one == b"Command Reader"


def test_passing_in_hook_deprecation():
    # Given a mock serial with 2 hooks, one of them custom
    @register_hook(hook_type_enum="post_reading", hooked_classes=[CannedQueries])
    def hook(self, hooked, result, data, **kwargs):
        if "1" in data:
            return "hook"
        return result

    approach_hook = ApproachHook(attributes={"temp"}, include_or_exclude="include")

    mock = MockSerial(config_key="deprecated", config_path=CONFIG_PATH_DEPRECATIONS, hooks=[approach_hook, hook])

    # When we set the temperature higher (not instantly since it is doesn't exclude temp)
    query_device(mock, "set -temp 25")
    # and then query it repeatedly
    old_value = float(query_device(mock, "get -temp")[:-2])
    for _ in range(5):
        time.sleep(0.001)
        # we grab the numeric part of the temp, leaving behind the write terminator
        new_value = float(query_device(mock, "get -temp")[:-2])

        # then each new value is strictly larger than the old value
        assert new_value > old_value

    # And when we query something with a 1
    hk = query_device(mock, "1")

    # Then it should trigger the custom hook and return "hook"
    assert hk == b"hook"



import abc
import functools
import logging
import os
import pathlib
import sys
import warnings
from builtins import bytes
from builtins import (
    str as unicode,  # python 3 doesn't have a unicode type, and we don't want to override python2 str type
)
from collections import OrderedDict
from datetime import datetime

import pandas as pd
import pkg_resources

logger = logging.getLogger(__name__)

IS_PYTHON3 = sys.version_info >= (3, 0, 0)

PYSERIAL_FIELDS_TO_EXCLUDE = {"BAUDRATES", "BYTESIZES", "STOPBITS", "PARITIES", "portstr", "writeTimeout"}

SENTINEL = object()

# making ABCs more uniform and compatible between python 2 and 3
ABC = abc.ABCMeta("ABC", (object,), {})


def add_created_at(init):
    """
    init decorator that adds `_created_at` attribute. May add future functionality later"""

    @functools.wraps(init)
    def wrapper(self, *args, **kwargs):
        init(self, *args, **kwargs)
        self._created_at = datetime.now()

    return wrapper


def fixpath(path):
    r"""
    On a linux like system (Posix system), it can't handle windows file pathing,
    this is a way to resolve that.
    On a linux system it first converts it to a windows path (where it handles both \ and /, thus
    reading it as a valid path), and then when it is a valid path, convert it back to a linux path.
    On a windows system, it is just happy to read it as a pathlib.Path (which will use either a
    WindowsPath on windows systems) since \, /, and \\ are fine.

    We then make sure to return a string.
    """
    if os.name == "posix":
        new_path = pathlib.PureWindowsPath(path).as_posix()
    else:
        new_path = pathlib.Path(path)
    return str(new_path)


def deunicodify_hook(pairs):
    """Hook for ``json.load`` to convert unicode json to byte json for python 2"""
    new_pairs = []
    for key, value in pairs:
        if isinstance(value, list):
            value = [val.encode("utf-8") if isinstance(val, unicode) else val for val in value]
        if isinstance(value, unicode):
            value = value.encode("utf-8")
        if isinstance(key, unicode):
            key = key.encode("utf-8")
        new_pairs.append((key, value))
    return OrderedDict(new_pairs)


def decode_bytes(byte_string):
    """Helper function to decode a bytestring"""
    if isinstance(byte_string, bytes):
        if IS_PYTHON3:
            return byte_string.decode("unicode_escape")
        else:
            return byte_string.decode("string_escape")
    raise TypeError("unicode strings are not supported, only use byte strings: {!r}".format(byte_string))


def encode_to_bytes(string, encoding="ascii"):
    """Helper function to encode a string as a bytestring"""
    if not isinstance(string, bytes):
        string = string.encode(encoding)
    return string


def decode_escape_char(string):
    """Helper function to decode escape character from either a string or a byte string"""
    if IS_PYTHON3:
        return bytes(string, "utf-8").decode("unicode_escape")
    else:
        return string.decode("string_escape")


def encode_escape_char(string):
    """Helper function to encode escape characters and non-unicode for printing to csv"""

    if IS_PYTHON3 or isinstance(string, unicode):
        return unicode(string.encode("unicode_escape"), "utf-8")
    else:
        return string.encode("string_escape")


def load_serial_df(path):
    """
    Load CSV into DataFrame.

    Necessary columns -> cmd, response
    """

    df = pd.read_csv(path, converters=dict(cmd=decode_escape_char, response=decode_escape_char), skipinitialspace=True)
    return df


def check_min_package_version(package, minimum_version, should_trunc_to_same_len=True):
    """Helper to decide if the package you are using meets minimum version requirement for some feature."""
    real_version = pkg_resources.get_distribution(package).version
    if should_trunc_to_same_len:
        minimum_version = minimum_version[0 : len(real_version)]

    logger.debug(
        "package %s, version: %s, minimum version to run certain features: %s", package, real_version, minimum_version
    )

    return real_version >= minimum_version


def get_path(path):
    """
    Helper function that if you pass in a path to a file or directory, returns the absolute path.

    Args:
        path (str): relative or absolute path to file or directory

    Returns:
        str: absolute path to file or directory
    """
    return os.path.abspath(path)


def make_path(path):
    """Make a path if the path does not already exist"""
    dir_name = os.path.dirname(path)
    if dir_name and not os.path.isdir(dir_name):
        os.makedirs(dir_name)


def int_to_char(int_):
    """Return an ascii character in byte string form for a given int"""
    return bytes([int_])


def is_terminated_with(input, terminator):
    """
    Helper to check if a given input has ends with the given terminator

    Args:
        input(str): the input to check
        terminator(str): some termination sequence

    Returns:
        bool : whether the input ends with the terminator
    """
    if len(input) < len(terminator):
        return False
    else:
        return input[-len(terminator) :] == terminator


def _get_subclasses(cls):
    """Get subclasses of passed in class and return a dictionary of them"""
    return {subclass.__name__: subclass for subclass in cls.__subclasses__()}


def deprecation(message, version_to_remove=None):
    warnings.simplefilter("always", DeprecationWarning)
    if version_to_remove:
        message += "\nScheduled to be removed on version {version}".format(version=version_to_remove)
    warnings.warn(message, DeprecationWarning)



import datetime
import time
from math import pi

import attr
import numpy as np

from granola.command_readers import CannedQueries, GettersAndSetters
from granola.enums import HookTypes, SetRelationship
from granola.hooks.base_hook import BaseHook, register_hook
from granola.utils import SENTINEL


@attr.s
class _ApproachHookAttributes(object):
    start_value = attr.ib(type=str, default=None, converter=float)
    end_value = attr.ib(type=str, default=None, converter=float)
    set_time = attr.ib(type=datetime.datetime, default=None)
    transition_time = attr.ib(type=int, default=None)


class ApproachHook(BaseHook):
    """
    Hook that will on applicable attributes, when a new value is set, it will approach that
    value over a period time (dictated by `transition_asc_scaling` and `transition_dsc_scaling`).
    It uses a hyperbolic tangent function to allow smooth transition from start and end values.

    Args:
        attributes (set, optional): getter and setter attributes to include or exclude from
            this hook.
        include_or_exclude (SetRelationship or str, optional): Whether to include `attributes` in
            the hook or exclude `attributes` from the hook.
        transition_asc_scaling (float): An increase of delta (change) of 1 unit takes this many seconds
            longer to transition. Must be
            positive.
        transition_dsc_scaling (float): A decrease of delta (change) of 1 unit takes this many seconds
            longer to transition. Must be
            negative.
    """

    hooked_classes = [GettersAndSetters]

    def __init__(
        self,
        attributes=None,
        include_or_exclude=SetRelationship.exclude,
        transition_asc_scaling=40,
        transition_dsc_scaling=100,
    ):
        super(ApproachHook, self).__init__(attributes=attributes, include_or_exclude=include_or_exclude)
        self.transition_asc_scaling = transition_asc_scaling
        self.transition_dsc_scaling = transition_dsc_scaling

        if self.transition_asc_scaling <= 0 or self.transition_dsc_scaling <= 0:
            raise ValueError("Inappropriate transition scaling value!")

    def pre_reading(self, hooked, data, **kwargs):
        """
        On applicable attributes, record meta data about the pre_reading value, the end_value,
        how long it should take to complete the transition, and when the transition started. Store
        all of this information in a meta data class.

        Args:
            hooked (GettersAndSetters): instance of hooked class
            data (str): Serial command
        """
        regex_match, response = hooked._get_matching_setter(data)
        if response:
            self._process_setter_approach_hook_helper(hooked, regex_match, hooked.attribute_vals)
        return data

    def post_reading(self, hooked, result, data, **kwargs):
        """
        On applicable attributes, retrieve meta data class information set from pre_reading hook,
        and based on that information (how long it has been since the set time, how long the
        transition is supposed to take, etc), calculate the new attribute value and return it.

        Args:
            hooked (GettersAndSetters): instance of hooked class
            result (str): Returned result from serial command.
            data (str): Serial command
        """
        if data in hooked.getters:
            attribute_vals = hooked.attribute_vals
            for attribute in attribute_vals:
                attrib = hooked.instrument_attributes[attribute].meta_data.get(_ApproachHookAttributes.__name__)
                if attrib:
                    seconds_ran = time.time() - attrib.set_time
                    value = attrib.start_value + 0.5 * (1 + np.tanh(6 * seconds_ran / attrib.transition_time - pi)) * (
                        attrib.end_value - attrib.start_value
                    )
                    attribute_vals[attribute] = value

            result = hooked.render_template(hooked.getters[data], attribute_vals)
        return result

    def _process_setter_approach_hook_helper(self, hooked, regex_match, attributes):
        """
        we can extract the match group by name because we gave
        them a name earlier for ease of book keeping now
        """
        for attribute in attributes:
            try:  # check for all attributes, only update attribute that matches regex
                end_value = regex_match.group(attribute)
            except IndexError:
                continue
            in_attribs = attribute in self.attributes
            include_attribs = self.include_or_exclude == SetRelationship.include.name
            exclude_attribs = self.include_or_exclude == SetRelationship.exclude.name
            if (in_attribs and include_attribs) or (not in_attribs and exclude_attribs):

                self.validate_attribute_type(hooked, attribute)

                attrib = _ApproachHookAttributes(
                    start_value=hooked.instrument_attributes[attribute].value, end_value=end_value, set_time=time.time()
                )

                delta = attrib.end_value - attrib.start_value
                transition_time = (
                    delta * self.transition_asc_scaling if delta >= 0 else delta * self.transition_dsc_scaling
                )
                attrib.transition_time = transition_time

                hooked.instrument_attributes[attribute].meta_data[_ApproachHookAttributes.__name__] = attrib
            hooked.instrument_attributes[attribute].value = end_value

    def validate_attribute_type(self, hooked, attribute):
        try:
            float(hooked.instrument_attributes[attribute].value)
        except ValueError:
            raise ValueError(
                "ApproachHook can only be applied with a float attribute.\nproblem attribute: %s" % attribute
            )


@register_hook(hook_type_enum=HookTypes.post_reading, hooked_classes=[CannedQueries])
def LoopCannedQueries(hooked, result, data, **kwargs):
    """
    Loop through canned queries to run when the returned query is SENTINEL, meaning
    the CannedQuery generator reached the end.

    Args:
        hooked (CannedQueries): CannedQueries instance to run the hook on.
        result (str): Passed in result to potentially modify.
        data (str): Serial command.

    Returns:
        str: the modified or unmodified serial command result

        SENTINEL (object): if the hook is passed in object, and it not acting on this
        serial command, it will just return SENTINEL unmodified.
    """
    if result is SENTINEL:
        hooked._start_serial_generator(data)
        result = next(hooked.serial_generator[data])
        return result
    return result


@register_hook(hook_type_enum=HookTypes.post_reading, hooked_classes=[CannedQueries])
def StickCannedQueries(hooked, result, **kwargs):
    """
    Stick on the last return query when the returned query is SENTINEL, meaning
    the CannedQuery generator reached the end.

    Args:
        hooked (CannedQueries): CannedQueries instance to run the hook on.
        result (str): Passed in result to potentially modify.

    Returns:
        str: the modified or unmodified serial command result
    """
    if result is SENTINEL:
        result = getattr(hooked, "last_reading", None)
        return result
    hooked.last_reading = result
    return result


HOOKS = {cls.__name__: cls for cls in BaseHook.__subclasses__()}


__doc__ = """
Hooks provide a way to inject customizable behavior at preditermined locations.

All hooks are subclasses of :class:`~granola.hooks.base_hook.BaseHook`. There are two ways to construct
a hook. Directly as a subclass of :class:`~granola.hooks.base_hook.BaseHook` or as a
function decorated with the :func:`~granola.hooks.base_hook.register_hook` decorator that converts a
simple function into the more full featured hook class.

For a detailed comparison of constructing a hook class versus a hook by decorator

.. seealso::

    Module :mod:`Hook API <granola.hooks.base_hook>`
"""



import inspect
import json
import logging
import os
from collections import OrderedDict

from serial import Serial

from granola.command_readers import BaseCommandReaders, CannedQueries, GettersAndSetters
from granola.hooks.base_hook import (
    BaseHook,
    _run_post_reading_hooks,
    _run_pre_reading_hooks,
)
from granola.utils import (
    IS_PYTHON3,
    SENTINEL,
    _get_subclasses,
    add_created_at,
    check_min_package_version,
    decode_bytes,
    deunicodify_hook,
    encode_to_bytes,
    fixpath,
    get_path,
    is_terminated_with,
)

try:
    from serial import PortNotOpenError  # for newer versions of pyserial
except ImportError:  # pragma: no cover
    PortNotOpenError = ValueError  # for pyserial 2.6


logger = logging.getLogger(__name__)


class Cereal(Serial):
    r"""
    Mock :std:doc:`Pyserial's <pyserial:index>` :class:`Serial <pyserial:serial.Serial>` class
    that is capable of mocking serial commands a responses to for both simple and sophisticated
    neeeds. Breakfast Cereal allows you to defined simple
    :class:`~granola.command_readersCannedQueries` that are are predefined
    serial commands and responses for various commands. As well as more complicated
    :class:`~granola.command_readers.GettersAndSetters` which allow on the fly changing
    the state of stored attributes. It also allows customizing the behavior of how
    individual or any set of commands are processed through :mod:`~granola.hooks.hooks`.

    Mock serials initialization follows a 2 step process. The first is with the normal ``__init__``
    method where you pass arguments to Cereal, you can do this before your project is going,
    as part of the setup. At this point, Cereal will be able to mock but it won't have many
    pyserial Serial attributes defined such as port, baudrate. Cereal is
    then able to be "initialized" a second time, following the signature of pyserial Serial class.
    This allows you to setup Cereal with its configuration before and then not have to
    change your code that initializes Pyserial much or at all through means of depencency injection
    or patching.

    The CSVs must have the columns cmd and response in them, it can have other columns as well.

    Args:
        command_readers (dict[BaseCommandReaders|str] | list[BaseCommandReaders|str], optional):
            Dictionary or list of :mod:`Command Readers <granola.command_readers>`. Command Readers
            can be passed in as a string representation, or you may pass in the class itself.
            Passing in a dictionary allows specifying your initialization options as well.
            Using a list allows you to pass in already initialized Command Reader, or a not
            initialized Command Reader that will be initialized to default values.
            You can pass in custom Command Readers as well. See
            :ref:`Custom Command Readers and Hooks Configuration`.
            defaults to ``(GettersAndSetters(), CannedQueries())``

        hooks (dict[BaseHook|str] | list[BaseHook|str], optional):
            Dictionary or list of :mod:`Hooks <granola.hooks.hook>`. Hooks
            can be passed in as a string representation, or you may pass in the class itself.
            Passing in a dictionary allows specifying your initialization options as well.
            Using a list allows you to pass in already initialized Hook, or a not
            initialized Hook that will be initialized to default values.
            You can pass in custom Hooks as well. See
            :ref:`Custom Command Readers and Hooks Configuration`.
            defaults to ``[]``

        data_path_root (str | Path, optional): Path to the where all data file paths (for
            Command Readers or Custom Hooks) will be referenced from.

            .. todo::

                probably just make this a thing inside the needed command readers

        unsupported_response(str, optional):
            The response returned for any command that does not have a defined response.
            Defaults to "Unsupported\r>"

        encoding(str, optional): The encoding scheme used to encode the serial commands and responses
            Defaults to "ascii"

    See Also
    --------
    :meth:`.mock_from_json` : Constructor from external configuration file

    :meth:`.BaseCommandReaders.get_reading` : Base Command Reader that defines the interface for other Command Readers

    :ref:`Configuration Overview` : Configuration Instructions and

    :ref:`Basic Overview of Mock Cereal and API` : Intro Tutorial

    Examples
    --------
    Initialize Breakfast Cereal with CannedQueries and GettersAndSetters Command Readers

    >>> command_readers = {
    ...     "CannedQueries": {"data": [{"1\r": "1", "2\r": ["2a", "2b"]}]},
    ...     "GettersAndSetters": {
    ...         "default_values": {"sn": "42"},  # We first initialize a default
    ...         "getters": [{"cmd": "get sn\r", "response": "{{ sn }}\r>"}],  # we define default
    ...         "setters": [{"cmd": "set sn {{ sn }}\r", "response": "OK\r>"}],  # and we define a setter
    ...     },
    ... }
    >>> cereal = Cereal(command_readers=command_readers)

    CannedQueries returns the correct response and by default it loops when it reaches the end of the defined responses

    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2a'
    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2b'
    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2a'

    Our GettersAndSetters initializes values to the defaults we specified, and then we can set them to new values

    >>> cereal.write(b"get sn\r")
    7
    >>> cereal.read(cereal.in_waiting)
    b'42\r>'
    >>> cereal.write(b"set sn our fancy new sn\r")  # if we don't read in between commands, it clears the buffer
    24
    >>> cereal.write(b"get sn\r")
    7
    >>> cereal.read(cereal.in_waiting)
    b'our fancy new sn\r>'

    If we send a command that is not defined, it sends back an Unsupported response, which is configurable

    >>> cereal.write(b"nonesense\r")
    10
    >>> cereal.read(cereal.in_waiting)
    b'Unsupported\r>'

    We can change the behavior of certain commands by adding :mod:`Hooks <granola.hooks.hooks>`.
    The :class:`Hooks <granola.hooks.hooks.StickCannedQueries>` causes the last defined response to not loop, but to
    stick on it and just repeat.

    >>> from granola import StickCannedQueries
    >>> hooks = {"StickCannedQueries": {"attributes": ["2\r"], "include_or_exclude": "include"}}
    >>> cereal = Cereal(command_readers=command_readers, hooks=hooks)
    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2a'
    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2b'
    >>> cereal.write(b"2\r")  # You'll notice here that it doesn't loop
    2
    >>> cereal.read(cereal.in_waiting)  # It just repeats the last response (it sticks)
    b'2b'
    >>> cereal.write(b"2\r")
    2
    >>> cereal.read(cereal.in_waiting)
    b'2b'
    """

    @add_created_at
    def __init__(
        self,
        command_readers=None,
        hooks=None,
        data_path_root=None,
        unsupported_response="Unsupported\r>",
        write_terminator="\r",
        encoding="ascii",
    ):
        self._data_path_root = (
            data_path_root if data_path_root is not None else os.path.join(os.getcwd(), "config.json")
        )
        hooks = hooks if hooks is not None else []
        command_readers = command_readers if command_readers is not None else []

        self._unsupported_response = unsupported_response
        self._encoding = encoding
        self._write_terminator = write_terminator

        self._readers_ = self._setup_command_readers_and_hooks(command_readers, hooks)

        self._hooks_ = []
        self._next_read = ""  # The next read for this "serial" device
        self._next_write = ""  # The current write buffer to the serial device

        if check_min_package_version("pyserial", "3.0"):  # mocking pyserial is_open/_isOpen for different pyserials
            self.is_open = True
        else:
            self._isOpen = True

    @classmethod
    def mock_from_json(cls, config_key, config_path="config.json", **kwargs):
        """
        Load configuration dictionary based on config_key and config_path and return a
        Breakfast Cereal class.

        If you don't pass in where config_path is, it will default to "config.json" in the
            root directory of your python path.
        """
        config = cls._load_json_config(config_key=config_key, config_path=config_path)
        kwargs.update(config)
        if "data_path_root" not in kwargs:
            kwargs["data_path_root"] = config_path
        return cls(**kwargs)

    @classmethod
    def _load_json_config(cls, config_key, config_path="config.json"):
        path = get_path(config_path)
        config_path = fixpath(path)

        logger.debug("%s config path: %s", cls.__name__, config_path)
        logger.debug("%s config key: %s", cls.__name__, config_key)

        with open(config_path) as f:
            if IS_PYTHON3:
                config = json.load(f, object_pairs_hook=OrderedDict)
            else:
                config = json.load(f, object_pairs_hook=deunicodify_hook)
        config = config[config_key]
        return config

    def __call__(self, *args, **kwargs):
        """
        Method to initialize pyserial Serial object. Serperating out the initialization of
        Breakfast Cereal into its own ``__init__`` method and then initializing pyserial in its
        __call__ method allows you to setup up Breakfast Cereal configuration outside of your
        code, and then have your code initialize it again to initialize the pyserial part of
        Breakfast Cereal. That way you can modify your own code as minimally as possible.

        Args:
            args: args to pass to pyserial
            kwargs: keyword arguments to pass to pyserial

        Raises:
            TypeError if you use __call__ more than once (and therefore pyserial is already
            initialized).
        """
        if not hasattr(self, "_port"):

            logger.debug("%s initializing Mock Serial with Serial args %s and kwargs %s", self, args, kwargs)

            is_open = self._is_open
            super(Serial, self).__init__(*args, **kwargs)
            self._is_open = is_open  # reset the port back to open since pyserial will close it on a dummy mock port
        else:
            raise TypeError("{cls} object is not callable".format(cls=self.__class__))

        return self

    def __str__(self):
        port = getattr(self, "port", "")
        port_str = " on %s" % port if port else ""
        return "{device}{port}".format(device=self.__class__.__name__, port=port_str)

    def read(self, size=1):
        """Mock :meth:`pyserial:serial.Serial.read`. Return number of bytes in self._next_read based on `size`"""
        read = bytearray()
        if size > 0:
            self._next_read = encode_to_bytes(self._next_read, self._encoding)
            while len(read) < size and len(self._next_read):
                read.append(self._next_read[0])
                self._next_read = self._next_read[1:]
        read = bytes(read)

        logger.info("%s read: %r", self, read)

        return read

    def write(self, data):
        """
        Mock :meth:`pyserial:serial.Serial.write` by seeding a serial command generator

        Args:
            data (byte_str): serial command
        """
        logger.info("%s write: %r", self, data)

        self._verify_open()

        data = decode_bytes(data)

        for d in data:
            self._next_write += d
            if is_terminated_with(self._next_write, self._write_terminator):
                # Based on observation, writes with multiple carriage returns only return the result up to the first
                # So data after the first terminator is basically ignored

                next_read = None
                _run_pre_reading_hooks(hooked=self, data=self._next_write)

                for reader in self._readers_.values():
                    next_read = reader.get_reading(data=self._next_write)
                    if next_read is not None:
                        self._next_read = next_read
                        break

                if next_read is None or next_read is SENTINEL:
                    self._next_read = self._unsupported_response
                    # If a response is not handled by the hooks and returns SENTINEL, return unsupported with warning
                    if next_read is SENTINEL:

                        logger.warning("%s unhandled response return from hooks. Defaulting to Unsupported Response!")

                self._next_write = ""  # once we grab the next read, clear the next write
        self._next_read = _run_post_reading_hooks(hooked=self, result=self._next_read, data=self._next_write)
        return len(data)

    if check_min_package_version("pyserial", "3.0"):

        def reset_input_buffer(self):
            """
            A wrapper for serial.reset_input_buffer that also clears the current read buffer.
            Should only be used with pyserial versions >= 3.0"""
            self._clear_input()

        def reset_output_buffer(self):
            """
            A wrapper for serial.reset_input_buffer that also clears the current write buffer.
            Should only be used with pyserial versions >= 3.0"""
            self._clear_output()

        @property
        def in_waiting(self):
            """mocking pyserial's in_waiting"""
            return self._in_waiting

        @property
        def out_waiting(self):
            """mocking pyserial's out_waiting"""
            return self._out_waiting

    else:

        def flushInput(self):
            """
            A wrapper for serial.FlushInput that also clears the current read buffer.
            Should only be used with pyserial versions <= 3.0"""
            self._clear_input()

        def flushOutput(self):
            """
            A wrapper for serial.FlushOutput that also clears the current write buffer.
            Should only be used with pyserial versions <= 3.0"""
            self._clear_output()

        def inWaiting(self):
            """mocking pyserial's inWaiting"""
            return self._in_waiting

        def outWaiting(self):
            """mocking pyserial's outWaiting"""
            return self._out_waiting

    @property
    def _is_open(self):
        """internal is_open so we always return the right version depending on what pyserial we have"""
        if check_min_package_version("pyserial", "3.0"):
            return self.is_open
        else:
            return self._isOpen

    @_is_open.setter
    def _is_open(self, value):
        """internal is_open so we always return the right version depending on what pyserial we have"""
        if check_min_package_version("pyserial", "3.0"):
            self.is_open = value
        else:
            self._isOpen = value

    def close(self):
        self._is_open = False

    def open(self):  # TODO madeline raise SerialException error if _port is none or if already open
        self._is_open = True

    def _clear_input(self):
        self._next_read = ""

    def _clear_output(self):
        self._next_write = ""

    @property
    def _in_waiting(self):
        return len(self._next_read)

    @property
    def _out_waiting(self):
        return 0  # Based on my observations, devices clear this buffer instantly

    def _verify_open(self):
        """
        verify if the `port` is open

        Raises:
            PortNotOpenError
        """
        if not self._is_open:
            raise PortNotOpenError

    def _setup_command_readers_and_hooks(self, command_readers, hooks):
        readers = OrderedDict()
        config_command_readers = self._read_file_config_command_readers(command_readers)
        if config_command_readers:
            for reader in config_command_readers:
                readers[reader.__class__.__name__] = reader
        else:
            command_readers = (GettersAndSetters, CannedQueries)
            for reader in command_readers:
                cr = reader()
                readers[cr.__class__.__name__] = cr

        config_hooks = self._read_file_config_hooks(hooks)
        if config_hooks:
            hooks = config_hooks
        for hook in hooks:
            if inspect.isclass(hook):  # initialize uninitialized hook to their default args
                hook = hook()
            for cls in hook.hooked_classes:
                readers[cls.__name__].register_hook(hook)

        for reader in readers.values():
            reader.assign_default_hook()

        return readers

    def _read_file_config_from_classes(baseclass):
        """
        closure to generate functions that will match passed in configuration option classes
        to a subclass of ``baseclass``. This allows users to pass in classes they want to use
        for things like Hooks, CommandReaders or other things inside a configuration dictionary (JSON).

        Args:
            baseclass (class): Class to check all subclasses of when you know class from config_options
        """

        def _inner(self, config_options):
            subclasses = _get_subclasses(baseclass)
            classes = []
            if isinstance(config_options, list):
                for cls in config_options:
                    if isinstance(cls, str):
                        c = subclasses[cls]()
                    elif inspect.isclass(cls):  # initialize uninitialized class
                        c = cls()
                    else:
                        c = cls
                    classes.append(c)
            else:
                for cls, options in config_options.items():
                    # add config path, but will be overwritten if defined in options
                    opts = {"data_path_root": self._data_path_root}
                    opts.update(options)
                    if isinstance(cls, str):
                        c = subclasses[cls](**opts)
                    elif inspect.isclass(cls):  # initialize uninitialized class
                        c = cls(**opts)
                    else:
                        c = cls
                    classes.append(c)
            return classes

        return _inner

    _read_file_config_hooks = _read_file_config_from_classes(BaseHook)

    _read_file_config_command_readers = _read_file_config_from_classes(BaseCommandReaders)


__doc__ = """
This module provides the core of GRANOLA with the Cereal class (affectionately called Breakfast Cereal).
Breakfast Cereal is the core Mock Serial object that handles mocking pyserial's Serial class. It does
this with a number of :mod:`Command Readers <granola.command_readers>` (:ref:`Guide <Configuration Overview>`)
that each are responsible for processing a number of serial commands in different ways. This allows
very extensible functionality with different types of serial commands, and user specific customization
by defining your own Command Readers.

You can further customize the behavior of Command Readers by adding in :mod:`Hook <granola.hooks.hooks>`
that run a preditermined spots. Also see :ref:`Configuration Overview`.
"""



import abc
import copy
import inspect
import logging
import os
import random
import re
from collections import OrderedDict
from pathlib import Path

import jinja2
import jinja2.meta
import pandas as pd

import granola.hooks
from granola.enums import RandomizeResponse, get_attribute_from_enum, validate_enum
from granola.hooks.base_hook import wrap_in_hooks
from granola.utils import ABC, IS_PYTHON3, SENTINEL, fixpath, load_serial_df

logger = logging.getLogger(__name__)


class SerialCmds(object):
    """
    Serial Command DataFrame container object. Required columns for DataFrame are `cmd` and `response`,
    Other columns are optional, and may or may not be used by certain Command Readers.

    Args:
        data (list[pd.DataFrame]): list of DataFrames containing serial command, response pairs,
            as well potentially other columns.
        will_randomize_responses (RandomizeResponse): Whether the responses of each command will be
            randomized or not.

    See Also:

        Class :class:`~granola.command_readers.CannedQueries`

        Config guide :ref:`Canned Queries Configuration`
    """

    def __init__(self, data=None, will_randomize_responses=RandomizeResponse.not_randomized):
        self.data = data if data is not None else []  # type: list[pd.DataFrame]
        self.will_randomize_responses = get_attribute_from_enum(will_randomize_responses, "name")
        validate_enum(self.will_randomize_responses, RandomizeResponse)

    def add_dataframe(self, df):
        """Add dataframe to data"""
        self.data.append(df)

    def add_df_from_file(self, file, data_path_root=None, **kwargs):
        """
        Add a df to data from a csv file path

        Args:
            file (str): Path to csv file.
            data_path_root (str, optional): Path to configuration. Required if file path is not an absolute path
            extra_fields (dict, optional): Dictionary of extra fields to add to DataFrame of commands. Either
                key will be mapped to a new column in the DataFrame, and each value can either be
                a single value, in which case it will be broadcast to all rows. Or a list of values
                the same length as the number of rows in the data.
        """
        file = str(file)
        if not os.path.isabs(file):
            if data_path_root is None:
                raise TypeError("`data_path_root` must be specified when using a relative file path")
            data_path_root = str(data_path_root)
            file = os.path.join(os.path.dirname(data_path_root), file)
        file = fixpath(file)
        logger.debug("Canned query path %s: ", file)
        df = load_serial_df(file)
        if kwargs:
            self._append_extra_fields_to_df(df, **kwargs)
        return self.add_dataframe(df)

    def add_df_from_dict(self, dic, **kwargs):
        """
        Add a df to data from a dictionary of serial commands. The format for this dictionary
        can take a few forms. It can either be expressed in a JSON configuration or passed as a python dictionary
        to Cereal. Here is the most basic form, where each command is mapped directly to a single response.

        Args:
            dic (dict): Dictionary of serial commands. See examples above for formatting.
            extra_fields (dict): Dictionary of extra fields to add to DataFrame of commands. Either
                key will be mapped to a new column in the DataFrame, and each value can either be
                a single value, in which case it will be broadcast to all rows. Or a list of values
                the same length as the number of rows in the data.
        """
        data = []
        d = copy.deepcopy(dic)
        for cmd, value in d.items():
            self._transverse_dict_and_append(data, cmd, value)
        df = pd.DataFrame.from_records(data)
        if kwargs:
            self._append_extra_fields_to_df(df, **kwargs)
        return self.add_dataframe(df)

    def _transverse_dict_and_append(self, data, cmd, value, **extra_fields):
        if isinstance(value, (str, int, float)):  # End of node
            self._append_to_list_of_rows(data, cmd, value, **extra_fields)
            return

        inner_extra_fields, response = self._extract_respose(value)

        if isinstance(response, (list, tuple)):
            for index, item in enumerate(response):
                if isinstance(item, (list, tuple)):
                    resp = item[0]
                    ex_fields = item[1]

                    intersection = set(ex_fields).intersection(inner_extra_fields)
                    if intersection:
                        for field in intersection:
                            scalar = isinstance(inner_extra_fields[field], (float, str, int))
                            if not scalar:
                                raise KeyError(
                                    "Extra Fields cannot be specified for both individual commands, and as a list."
                                    "\nEither specify a extra field for individual commands"
                                    " or as a list the same length as your command responses."
                                )

                    response_ = resp
                    extra_fields_ = {
                        key: val[index] if isinstance(val, (list, tuple)) else val
                        for key, val in inner_extra_fields.items()
                    }
                    extra_fields_.update(ex_fields)
                elif isinstance(item, (str, int, float)):
                    response_ = item
                    extra_fields_ = {
                        key: val[index] if isinstance(val, (list, tuple)) else val
                        for key, val in inner_extra_fields.items()
                    }
                else:
                    raise ValueError("Invalid JSON format for canned_queries")
                self._append_to_list_of_rows(data, cmd, response_, **extra_fields_)
        else:
            response_ = response
            extra_fields_ = inner_extra_fields
            self._append_to_list_of_rows(data, cmd, response_, **extra_fields_)

    @staticmethod
    def _extract_respose(value):
        if isinstance(value, dict):
            response = value.pop("response")
        elif isinstance(value, (list, tuple)):  # value must be a list for it to be valid json
            response = list(value)  # copy to make believe that it was a dict of form {response: []}
            value = {}
        else:
            raise ValueError("Invalid JSON format for canned_queries")
        return value, response

    @staticmethod
    def _append_to_list_of_rows(data, cmd, response, **ex_fields):
        data.append(dict(cmd=cmd, response=response, **ex_fields))

    @staticmethod
    def _append_extra_fields_to_df(data, **kwargs):
        for field, value in kwargs.items():
            if field in data.columns:
                data.loc[data[field].isna(), field] = value
            else:
                data[field] = value


class InstrumentAttribute(object):
    def __init__(self, name, value, meta_data=None):
        self.name = name
        self.value = value
        self.def_value = value
        self.meta_data = meta_data if meta_data is not None else OrderedDict()


class BaseCommandReaders(ABC):
    """
    BaseCommandReaders Class that sets the interface for other CommandReaders. The basic form of
    a command reader is that Cereal will interact with it solely through its ``__init__``
    and :meth:`granola.command_readers.BaseCommandReaders.get_reading`. The ``get_reading``
    method is the method that will be called, handed in a serial command, processed in some way,
    and returned. If the Command Reader can not process the serial command, it instead returns
    None or a SENTINEL object (more on that below) to indicate that the supplied serial command
    isn't defined for that Command Reader, allowing any other Command Readers to process it instead.

    The `get_reading` commands are also decorated with @wrap_in_hooks, which will run ``_hooks_``
    attributes (of type :mod:`~granola.hooks.hooks`) that have a defined ``prehook`` method before
    ``get_reading`` and ``_hooks_`` with a defined ``posthook`` method after after ``get_reading``.

    Because "" is a valid serial command response, we use None to signify an unsupported response.
    We also have a defined SENTINEL object for responses that are undefined in other ways.
    For example, Canned Queries (command and response defined from list) by default loop back when
    they get to the end of the list. However, if you exclude a command from that list, and then don't
    handle it another way, we return a SENTINEL object instead of having undefined behavior.
    Cereal will treat this as an undefined response, but also emit a warning log message
    as well.

    Args:
        hooks(list[BaseHook], optional): List of Hooks to run on this Command Reader.
            Defaults to `[]`
        data_path_root(str | Path, optional): Base path for all data file paths for this command reader
            (such as defining the serial commands). Not every Command
            Reader, or every way to initialize a certain Command Reader uses data files, so
            this doesn't apply to every Command Reader.
            Defaults to current working directory.

    See Also
    --------
    :class:`.Cereal` : Breakfast Cereal class used for mocking
    :ref:`Custom Command Readers and Hooks Configuration` : Configuration Overview
    :ref:`Basic Overview of Mock Cereal and API` : Intro Tutorial
    """

    def __init__(self, hooks=None, data_path_root=None, *args, **kwargs):
        super(BaseCommandReaders, self).__init__()
        self._hooks_ = hooks if hooks is not None else []
        self._data_path_root = data_path_root if data_path_root is not None else os.getcwd()

    @wrap_in_hooks
    @abc.abstractmethod
    def get_reading(self, data):
        """
        Processes incoming serial command data and return a response if a matching one is found.

        Args:
            data (str): Incoming serial command

        Returns:
            str | None | SENTINEL: the response from either the matching getter or setter.
                Returns None if serial command is not supported.
                Returns SENTINEL as a flag for some kind of none processing of the serial command
                by the hooks and CommandReader, dependent on the individual CommandReader.
        """

    def register_hook(self, hook):
        self._hooks_.append(hook)

    def assign_default_hook(self):
        """If self._hooks_ hooks is empty, add any default hooks for this Command Reader."""
        if not self._hooks_:
            pass


class GettersAndSetters(BaseCommandReaders):
    """
    Command Reader to handle getters and setters. It initializes the default values and stores those
    values in the attribute `self.instrument_attributes`. These attributes can then be modified with
    setters defined in the setters in `self.setters`, and then grabbed from the getters stored in
    `self.getters`.

    Inside the getters and setters, the formatting follows Jinja2 formatting syntax.

    Args:
        arguments for BaseCommandReaders

    See Also
    --------
    :ref:`Getters and Setters Configuration`

    :ref:`Advanced Getters and Setters`
    """

    def __init__(
        self,
        default_values=None,
        getters=None,
        setters=None,
        variable_start_string="{{",
        variable_end_string="}}",
        **kwargs
    ):
        super(GettersAndSetters, self).__init__(**kwargs)
        self._variable_start_string = variable_start_string
        self._variable_end_string = variable_end_string
        self.jinja_env = jinja2.Environment(
            variable_start_string=variable_start_string,
            variable_end_string=variable_end_string,
            loader=jinja2.BaseLoader(),
        )
        default_values = default_values if default_values is not None else OrderedDict()
        getters = getters if getters is not None else OrderedDict()
        setters = setters if setters is not None else OrderedDict()
        self.instrument_attributes = OrderedDict()
        self.getters = OrderedDict()
        self.setters = OrderedDict()
        self._load_getters_and_setters(default_values, getters, setters)

    @wrap_in_hooks
    def get_reading(self, data):
        """
        Processes incoming serial command data and sends it to `self._process_getter` to see if
        it is a getter and get its response, if not, sends it to `self._process_setter`,
        to see if it is a setter and then set the respective setter and get its response,
        if not, it returns None.

        Args:
            data (str): Incoming serial command

        Returns:
            str | None: the response from either the matching getter or setter. If no matching
                getter or setter, then it returns None.
        """
        next_read = self._process_getter(data)
        if next_read is not None:
            return next_read
        next_read = self._process_setter(data)
        if next_read is not None:
            return next_read
        return

    @property
    def attribute_vals(self):
        attribute_vals = {
            attribute: instrument_attribute.value
            for attribute, instrument_attribute in self.instrument_attributes.items()
        }
        return attribute_vals

    def render_template(self, string, attribute_vals=None):
        attribute_vals = attribute_vals if attribute_vals is not None else self.attribute_vals
        command = string.replace("\r", "_\\r_").replace("\n", "_\\n_")
        template = self.jinja_env.from_string(command)
        result = template.render(**attribute_vals).replace("_\\r_", "\r").replace("_\\n_", "\n")
        return result

    def _load_getters_and_setters(self, default_values, getters, setters):
        """Loads default values, getters and setters"""

        if not default_values:
            return
        self._initialize_default_instrument_attributes(default_values)
        self._initialize_getters(getters)
        self._initialize_setters(setters)

    def _initialize_default_instrument_attributes(self, default_values):
        """Loads default values into `self.instrument_attributes`"""
        for attribute, default_value in default_values.items():
            self.instrument_attributes[attribute] = InstrumentAttribute(name=attribute, value=default_value)

    def _initialize_getters(self, getters):
        """
        Loads all getters from getters_and_setters["getters"] into `self.getters`. If getter
        does not get an attribute in self.instrument_attributes, raise Value Error"""
        for getter in getters:
            self._check_attributes_are_valid("getters", getter)

            self.getters[getter["cmd"]] = getter["response"]

    def _initialize_setters(self, setters):
        """
        Loads all setters from getters_and_setters["setters"] into `self.setters`. If setter
        does not get an attribute in self.instrument_attributes, raise Value Error"""
        for setter in setters:
            self._check_attributes_are_valid("setters", setter)

            regex = "{variable_start}(.*?){variable_end}".format(
                variable_start=self._variable_start_string, variable_end=self._variable_end_string
            )
            setter_regex = re.sub(regex, self._format_match_group, setter["cmd"])
            self.setters[setter_regex] = setter["response"]

    def _format_match_group(self, match):
        """
        Get and format match group. Default formatters create named match groups to
        extract the match group by the name of the attribute later for easier book keeping."""
        group = match.group(0)
        group = (
            group.strip(self._variable_start_string)
            .strip(
                self._variable_end_string,
            )
            .strip()
        )
        group = "(?P<{group}>.*)".format(group=group)
        return group

    def _process_getter(self, data):
        """Process getter if data matches getter format"""
        if data in self.getters:
            next_read = self.render_template(self.getters[data])

            return next_read
        return

    def _process_setter(self, data):
        """Process setter if data matches a setter regex"""
        regex_match, response_template = self._get_matching_setter(data)
        if not response_template:
            return
        self._proccess_setter_helper(regex_match, self.attribute_vals)

        response = self.render_template(response_template)
        return response

    def _get_matching_setter(self, data):
        """Process setter if data matches a setter regex"""
        for cmd_regex, response in self.setters.items():
            regex_match = re.search(cmd_regex, data)
            if regex_match:
                return regex_match, response

        return None, None

    def _proccess_setter_helper(self, regex_match, attributes):
        # we can extract the match group by name because we gave
        # them a name earlier for ease of book keeping now
        for attribute in attributes:
            try:  # check for all attributes, only update attribute that matches regex
                updated_value = regex_match.group(attribute)
            except IndexError:
                continue

            logger.debug("Updating %s to %s", attribute, updated_value)

            self.instrument_attributes[attribute].value = updated_value

    def _check_attributes_are_valid(self, attribute_type, template_string):
        """Check every attribute in attributes is an attribute in `self.instrument_attributes`,
        else raise ValueError"""
        for template in template_string.values():
            parsed_content = self.jinja_env.parse(template)
            attributes = jinja2.meta.find_undeclared_variables(parsed_content)
            for attribute in attributes:
                if attribute not in self.instrument_attributes:
                    raise ValueError(
                        "{attribute_type} attribute {attribute} not found in default_values."
                        "\nMake sure you initialize all values!"
                        "\nProblem string: {template_string}".format(
                            attribute_type=attribute_type, attribute=attribute, template_string=template
                        )
                    )


class CannedQueries(BaseCommandReaders):
    """
    Helper class that processes canned queries (queries defined ahead of time from a list
    such as a csv or configuration dictionary).

    It stores a dictionary of :class:`~granola.command_readers.SerialCmds`, which represents
    canned queries as pandas DataFrames of the serial commands. And then as a serial command comes in,
    it is directed to the correct :class:`~granola.command_readers.SerialCmds` and then creates a slice of
    your DataFrame to get only the matching serial commands to iterate over.

    See Also
    --------
    :ref:`Canned Queries Configuration` for examples on configuration formatting.

    :ref:`Custom Command Readers and Hooks Configuration` : Command Readers and Hook Overviews
    """

    def __init__(self, data=None, data_path_root=None, **kwargs):

        super(CannedQueries, self).__init__(data_path_root=data_path_root, **kwargs)
        self.data = data if data is not None else OrderedDict()
        self.serial_df = pd.DataFrame(columns=["cmd", "response"])  # default empty df
        serial_cmd_files_kwargs = self._extract_serial_cmd_file_kw_from_config(kwargs)
        self.serial_cmd_file = SerialCmds(**serial_cmd_files_kwargs)
        self.serial_generator = OrderedDict()

        for maybe_file in self.data:
            if isinstance(maybe_file, (str, Path)):
                self.serial_cmd_file.add_df_from_file(maybe_file, self._data_path_root, **kwargs)
            elif isinstance(maybe_file, dict):
                self.serial_cmd_file.add_df_from_dict(maybe_file, **kwargs)

        self._seed_serial_dfs()

    def assign_default_hook(self):
        if not self._hooks_:
            self._hooks_ = [granola.hooks.hooks.LoopCannedQueries()]

    @wrap_in_hooks
    def get_reading(self, data):
        """
        Process incoming canned queries by extracting just the serial commands that match the
        incoming serial command and then create a generator of those matching serial commands
        and response to iterate through.

        Args:
            data (str): Incoming serial command

        Returns:
            str | None | SENTINEL: the response from matching serial df.
            If no matching matching command is found, then it returns None.
            If The generator is exhausted, and no Hook is activate to reseed the generator or
            alter the behavior in some other way, return SENTINEL.

        See Also
        --------
        :mod:`Hooks <granola.hooks.hooks>` : Hook API information

        :ref:`Canned Queries Configuration` for examples on configuration formatting.

        :ref:`Custom Command Readers and Hooks Configuration` : Command Readers and Hook Overviews
        """
        try:
            if data not in self.serial_generator:
                self._start_serial_generator(data)
            next_read = next(self.serial_generator[data], SENTINEL)
        except KeyError:
            next_read = None
        return next_read

    def _start_serial_generator(self, cmd):
        serial_df = self._filter_serial_df(cmd=cmd)  # type: pd.DataFrame
        if not serial_df.empty:
            serial_df_generator = self._get_generator_from_df(serial_df, self.serial_cmd_file.will_randomize_responses)
            self.serial_generator[cmd] = serial_df_generator

    def _extract_serial_cmd_file_kw_from_config(self, kw):
        """
        extract signature of SerialCmds and recursively search `self._config` for matching
        keys ad return all found key value pairs.
        """
        if IS_PYTHON3:  # pragma: no cover
            inspection_getargspec = inspect.getfullargspec
        else:
            inspection_getargspec = inspect.getargspec
        kwargs = OrderedDict()
        serial_cmd_args = inspection_getargspec(SerialCmds.__init__).args
        for arg in serial_cmd_args:
            if kw.get(arg) is not None:
                kwargs[arg] = kw[arg]
            if kw.get("canned_queries", {}).get(arg) is not None:
                kwargs[arg] = kw["canned_queries"][arg]
        logger.debug("%s extracted SerialCmds kwargs: %s", self, kwargs)
        return kwargs

    def _seed_serial_dfs(self):
        if self.serial_cmd_file.data:
            self.serial_df = pd.concat(objs=[df for df in self.serial_cmd_file.data])

    def _filter_serial_df(self, cmd):
        """
        Filter DataFrame to only the commands that match the inputted command

        This allows the CSVs to contain many different serial inputs, but it will only return
        the serial command that you pass in.

        Args:
            df (pd.DataFrame): DataFrame to filer.
            cmd (str): Command to filter against.

        Returns:
            pd.DataFrame: Filtered DataFrame.
        """
        df = self.serial_df
        conditions = df["cmd"] == cmd
        columns = ["response"]
        serial_df = df.loc[conditions, columns]
        return serial_df

    @staticmethod
    def _get_generator_from_df(df, will_randomize_responses):
        """Create a generator for the df so that when you call next on that generator, it
        gives you the next serial command each time, and not the first one. Allowing you to
        continue through the list of serial commands. Also gives the option to randomize results.

        Args:
            df (pd.DataFrame): DataFrame to iterate over
            serial_cmd_file (SerialCmds): SerialCmds to know if you are randomizing the results or just doing
                a straight iteration.
        """
        # TODO(madeline) only pass in randomize enum? Also clean up and make subfunctions?
        if will_randomize_responses == RandomizeResponse.not_randomized.name:
            for _, rows in df.iterrows():
                yield rows["response"]
        if (
            will_randomize_responses == RandomizeResponse.randomized_w_replacement.name
            or will_randomize_responses == RandomizeResponse.randomize_and_remove.name
        ):
            while df.shape[0] > 0:
                df_length = df.shape[0]
                row = random.randint(0, df_length - 1)
                yield df["response"].iloc[row]
                if will_randomize_responses == RandomizeResponse.randomize_and_remove.name:
                    df = df.drop(row)


__doc__ = """
Command Readers are the objects that handle the processing of individual serial
commands. Each serial command that comes in is processed by each Command Reader and
once a Command Reader returns a valid response, the next Command Readers will skip processing the command.
This allows you to change the behavior of individual commands by defining new Command Readers.
"""
